"""
This module contains all utility functions used throughout frads.
"""
from io import TextIOWrapper
import logging
import math
from pathlib import Path
import random
import string
import subprocess as sp
from typing import Any, Dict, Optional, List
from typing import Set
from typing import Tuple
from typing import Union

from frads import geom
from frads import parsers
from frads.types import Primitive
from frads.types import PaneRGB
from frads.types import ScatteringData
from frads.types import BSDFData


logger: logging.Logger = logging.getLogger("frads.utils")


GEOM_TYPE = ["polygon", "ring", "tube", "cone"]

MATERIAL_TYPE = ["plastic", "glass", "trans", "dielectric", "BSDF"]

BASIS_DICT = {
    "145": "Klems Full",
    "73": "Klems Half",
    "41": "Klems Quarter",
}

TREG_BASE = [
    (90.0, 0),
    (78.0, 30),
    (66.0, 30),
    (54.0, 24),
    (42.0, 24),
    (30.0, 18),
    (18.0, 12),
    (6.0, 6),
    (0.0, 1),
]

ABASE_LIST = {
    "Klems Full": [
        (0.0, 1),
        (5.0, 8),
        (15.0, 16),
        (25.0, 20),
        (35.0, 24),
        (45.0, 24),
        (55.0, 24),
        (65.0, 16),
        (75.0, 12),
        (90.0, 0),
    ],
    "Klems Half": [
        (0.0, 1),
        (6.5, 8),
        (19.5, 12),
        (32.5, 16),
        (46.5, 20),
        (61.5, 12),
        (76.5, 4),
        (90.0, 0),
    ],
    "Klems Quarter": [(0.0, 1), (9.0, 8), (27.0, 12), (46.0, 12), (66.0, 8), (90.0, 0)],
}


def tmit2tmis(tmit: float) -> float:
    """Convert from transmittance to transmissivity."""
    tmis = round(
        (math.sqrt(0.8402528435 + 0.0072522239 * tmit**2) - 0.9166530661)
        / 0.0036261119
        / tmit,
        3,
    )
    return max(0, min(tmis, 1))


def polygon2prim(polygon: geom.Polygon, modifier: str, identifier: str) -> Primitive:
    """Generate a primitive from a polygon."""
    return Primitive(modifier, "polygon", identifier, ["0"], polygon.to_real())


def unpack_idf(path: str) -> dict:
    """Read and parse and idf files."""
    with open(path, "r") as rdr:
        return parsers.parse_idf(rdr.read())


def sdata2bsdf(sdata: ScatteringData) -> BSDFData:
    """Convert sdata object to bsdf object."""
    basis = BASIS_DICT[str(sdata.ncolumn)]
    lambdas = angle_basis_coeff(basis)
    bsdf = []
    for irow in range(sdata.nrow):
        for icol, lam in zip(range(sdata.ncolumn), lambdas):
            bsdf.append(sdata.sdata[icol + irow * sdata.ncolumn] / lam)
    return BSDFData(bsdf, sdata.ncolumn, sdata.nrow)


def bsdf2sdata(bsdf: BSDFData) -> ScatteringData:
    """Covert a bsdf object into a sdata object."""
    basis = BASIS_DICT[str(bsdf.ncolumn)]
    lambdas = angle_basis_coeff(basis)
    sdata = []
    for irow in range(bsdf.nrow):
        for icol, lam in zip(range(bsdf.ncolumn), lambdas):
            sdata.append(bsdf.bsdf[icol + irow * bsdf.ncolumn] * lam)
    return ScatteringData(sdata, bsdf.ncolumn, bsdf.nrow)


def frange_inc(start, stop, step):
    """Generate increasing non-integer range."""
    r = start
    while r < stop:
        yield r
        r += step


def square2disk(in_square_a: float, in_square_b: float) -> tuple:
    """Shirley-Chiu square to disk mapping.
    Args:
        in_square_a: [-1, 1]
        in_square_b: [-1, 1]
    """
    if in_square_a + in_square_b > 0:
        if in_square_a > in_square_b:
            in_square_rgn = 0
        else:
            in_square_rgn = 1
    else:
        if in_square_b > in_square_a:
            in_square_rgn = 2
        else:
            in_square_rgn = 3
    out_disk_r = [in_square_a, in_square_b, -in_square_a, -in_square_b][in_square_rgn]
    if in_square_b * in_square_b > 0:
        phi_select_4 = 6 - in_square_a / in_square_b
    else:
        phi_select_4 = 0
    phi_select = [
        in_square_b / in_square_a,
        2 - in_square_a / in_square_b,
        4 + in_square_b / in_square_a,
        phi_select_4,
    ]
    out_disk_phi = math.pi / 4 * phi_select[in_square_rgn]
    out_disk_x = out_disk_r * math.cos(out_disk_phi)
    out_disk_y = out_disk_r * math.sin(out_disk_phi)
    return out_disk_x, out_disk_y, out_disk_r, out_disk_phi


def id_generator(size: int = 3, chars: Optional[str] = None) -> str:
    """Generate random characters."""
    if chars is None:
        chars = string.ascii_uppercase + string.digits
    return "".join(random.choice(chars) for _ in range(size))


def unpack_primitives(file: Union[str, Path, TextIOWrapper]) -> List[Primitive]:
    """Open a file a to parse primitive."""
    if isinstance(file, TextIOWrapper):
        lines = file.readlines()
    else:
        with open(file, "r", encoding="ascii") as rdr:
            lines = rdr.readlines()
    return parsers.parse_primitive(lines)


def primitive_normal(primitive_paths: List[str]) -> Set[geom.Vector]:
    """Return a set of normal vectors given a list of primitive paths."""
    _primitives: List[Primitive] = []
    _normals: List[geom.Vector]
    for path in primitive_paths:
        _primitives.extend(unpack_primitives(path))
    _normals = [parsers.parse_polygon(prim.real_arg).normal for prim in _primitives]
    return set(_normals)


def samp_dir(primlist: list) -> geom.Vector:
    """Calculate the primitives' average sampling
    direction weighted by area."""
    primlist = [p for p in primlist if p.ptype in ("polygon", "ring")]
    normal_area = geom.Vector()
    for prim in primlist:
        polygon = parsers.parse_polygon(prim.real_arg)
        normal_area += polygon.normal.scale(polygon.area)
    sdir = normal_area.scale(1.0 / len(primlist))
    sdir = sdir.normalize()
    return sdir


def up_vector(primitives: list) -> geom.Vector:
    """Define the up vector given primitives.

    Args:
        primitives: list of dictionary (primitives)

    Returns:
        returns a str as x,y,z

    """
    xaxis = geom.Vector(1, 0, 0)
    yaxis = geom.Vector(0, 1, 0)
    norm_dir = samp_dir(primitives)
    if norm_dir not in (xaxis, xaxis.scale(-1)):
        upvect = xaxis.cross(norm_dir)
    else:
        upvect = yaxis.cross(norm_dir)
    return upvect


def neutral_plastic_prim(
    mod: str, ident: str, refl: float, spec: float, rough: float
) -> Primitive:
    """Generate a neutral color plastic material.
    Args:
        mod(str): modifier to the primitive
        ident(str): identifier to the primitive
        refl (float): measured reflectance (0.0 - 1.0)
        specu (float): material specularity (0.0 - 1.0)
        rough (float): material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [spec, refl, rough]), err_msg
    real_args = [5, refl, refl, refl, spec, rough]
    return Primitive(mod, "plastic", ident, ["0"], real_args)


def neutral_trans_prim(
    mod: str, ident: str, trans: float, refl: float, spec: float, rough: float
) -> Primitive:
    """Generate a neutral color plastic material.
    Args:
        mod(str): modifier to the primitive
        ident(str): identifier to the primitive
        refl (float): measured reflectance (0.0 - 1.0)
        specu (float): material specularity (0.0 - 1.0)
        rough (float): material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    color = trans + refl
    t_diff = trans / color
    tspec = 0
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [spec, refl, rough]), err_msg
    real_args = [7, color, color, color, spec, rough, t_diff, tspec]
    return Primitive(mod, "trans", ident, ["0"], real_args)


def color_plastic_prim(mod, ident, refl, red, green, blue, specu, rough) -> Primitive:
    """Generate a colored plastic material.
    Args:
        mod(str): modifier to the primitive
        ident(str): identifier to the primitive
        refl (float): measured reflectance (0.0 - 1.0)
        red; green; blue (int): rgb values (0 - 255)
        specu (float): material specularity (0.0 - 1.0)
        rough (float): material roughness (0.0 - 1.0)

    Returns:
        A material primtive
    """
    err_msg = "reflectance, speculariy, and roughness have to be 0-1"
    assert all(0 <= i <= 1 for i in [specu, refl, rough]), err_msg
    red_eff = 0.3
    green_eff = 0.59
    blue_eff = 0.11
    weighted = red * red_eff + green * green_eff + blue * blue_eff
    matr = round(red / weighted * refl, 3)
    matg = round(green / weighted * refl, 3)
    matb = round(blue / weighted * refl, 3)
    real_args = [5, matr, matg, matb, specu, rough]
    return Primitive(mod, "plastic", ident, "0", real_args)


def glass_prim(
    mod, ident, tr: float, tg: float, tb: float, refrac: float = 1.52
) -> Primitive:
    """Generate a glass material.

    Args:
        mod (str): modifier to the primitive
        ident (str): identifier to the primtive
        tr, tg, tb (float): transmmisivity in each channel (0.0 - 1.0)
        refrac (float): refraction index (default=1.52)
    Returns:
        material primtive (dict)

    """
    tmsv_red = tmit2tmis(tr)
    tmsv_green = tmit2tmis(tg)
    tmsv_blue = tmit2tmis(tb)
    real_args = [4, tmsv_red, tmsv_green, tmsv_blue, refrac]
    return Primitive(mod, "glass", ident, ["0"], real_args)


def bsdf_prim(
    mod,
    ident,
    xmlpath,
    upvec,
    pe: bool = False,
    thickness: float = 0.0,
    xform=None,
    real_args: str = "0",
) -> Primitive:
    """Create a BSDF primtive."""
    str_args = [xmlpath, str(upvec)]
    str_args_count = 5
    if pe:
        _type = "aBSDF"
    else:
        str_args_count += 1
        str_args = [str(thickness), *str_args]
        _type = "BSDF"
    if xform is not None:
        str_args_count += len(xform.split())
        str_args.extend(*xform.split())
    else:
        str_args.append(".")
    str_args = [str(str_args_count), *str_args]
    return Primitive(mod, _type, ident, str_args, real_args)


def lambda_calc(theta_lr: float, theta_up: float, nphi: float) -> float:
    """."""
    return (
        (
            math.cos(math.pi / 180 * theta_lr) ** 2
            - math.cos(math.pi / 180 * theta_up) ** 2
        )
        * math.pi
        / nphi
    )


def angle_basis_coeff(basis: str) -> List[float]:
    """Calculate klems basis coefficient"""
    ablist = ABASE_LIST[basis]
    lambdas = []
    for i in range(len(ablist) - 1):
        tu = ablist[i + 1][0]
        tl = ablist[i][0]
        np = ablist[i][1]
        lambdas.extend([lambda_calc(tl, tu, np) for _ in range(np)])
    return lambdas


def opt2list(opt: dict) -> List[str]:
    """Convert option dictionary to list.

    Key: str
    Value: str | float | int | bool | list

    Args:
        opt: option dictionary
    Returns:
        A list of strings
    """
    out = []
    for key, value in opt.items():
        if isinstance(value, str):
            if key == "vf":
                out.extend(["-" + key, value])
            else:
                out.append(f"-{key}{value}")
        elif isinstance(value, bool):
            if value:
                out.append(f"-{key}+")
            else:
                out.append(f"-{key}-")
        elif isinstance(value, (int, float)):
            out.extend(["-" + key, str(value)])
        elif isinstance(value, list):
            out.extend(["-" + key, *map(str, value)])
    return out


def calc_reinsrc_dir(
    mf: int, x1: float = 0.5, x2: float = 0.5
) -> Tuple[List[geom.Vector], List[float]]:
    """
    Calculate Reinhart/Treganza sampling directions.
    Direct translation of Radiance reinsrc.cal file.

    Args:
        mf(int): multiplication factor.
        x1(float, optional): bin position 1
        x2(float, optional): bin position 2
    Returns:
        A list of geom.Vector
        A list of solid angle associated with each vector
    """

    def rnaz(r):
        """."""
        if r > (mf * 7 - 0.5):
            return 1
        return mf * TNAZ[int(math.floor((r + 0.5) / mf))]

    def raccum(r):
        """."""
        if r > 0.5:
            return rnaz(r - 1) + raccum(r - 1)
        return 0

    def rfindrow(r, rem):
        """."""
        if (rem - rnaz(r)) > 0.5:
            return rfindrow(r + 1, rem - rnaz(r))
        return r

    TNAZ = [30, 30, 24, 24, 18, 12, 6]
    rowMax = 7 * mf + 1
    runlen = 144 * mf**2 + 3
    rmax = raccum(rowMax)
    alpha = 90 / (mf * 7 + 0.5)
    dvecs = []
    omegas = []
    for rbin in range(1, runlen):
        rrow = rowMax - 1 if rbin > (rmax - 0.5) else rfindrow(0, rbin)
        rcol = rbin - raccum(rrow) - 1
        razi_width = 2 * math.pi / rnaz(rrow)
        rah = alpha * math.pi / 180
        razi = (rcol + x2 - 0.5) * razi_width if rbin > 0.5 else 2 * math.pi * x2
        ralt = (rrow + x1) * rah if rbin > 0.5 else math.asin(-x1)
        cos_alt = math.cos(ralt)
        if rmax - 0.5 > rbin:
            romega = razi_width * (math.sin(rah * (rrow + 1)) - math.sin(rah * rrow))
        else:
            romega = 2 * math.pi * (1 - math.cos(rah / 2))
        dx = math.sin(razi) * cos_alt
        dy = math.cos(razi) * cos_alt
        dz = math.sin(ralt)
        dvecs.append(geom.Vector(dx, dy, dz))
        omegas.append(romega)
    return dvecs, omegas


def pt_inclusion(pt: geom.Vector, polygon_pts: List[geom.Vector]) -> int:
    """Test whether a point is inside a polygon
    using winding number algorithm."""

    def isLeft(pt0, pt1, pt2):
        """Test whether a point is left to a line."""
        return (pt1.x - pt0.x) * (pt2.y - pt0.y) - (pt2.x - pt0.x) * (pt1.y - pt0.y)

    # Close the polygon for looping
    # polygon_pts.append(polygon_pts[0])
    polygon_pts = [*polygon_pts, polygon_pts[0]]
    wn = 0
    for i in range(len(polygon_pts) - 1):
        if polygon_pts[i].y <= pt.y:
            if polygon_pts[i + 1].y > pt.y:
                if isLeft(polygon_pts[i], polygon_pts[i + 1], pt) > 0:
                    wn += 1
        else:
            if polygon_pts[i + 1].y <= pt.y:
                if isLeft(polygon_pts[i], polygon_pts[i + 1], pt) < 0:
                    wn -= 1
    return wn


def gen_grid(polygon: geom.Polygon, height: float, spacing: float) -> list:
    """Generate a grid of points for orthogonal planar surfaces.

    Args:
        polygon: a polygon object
        height: points' distance from the surface in its normal direction
        spacing: distance between the grid points
    Returns:
        List of the points as list
    """
    vertices = polygon.vertices
    plane_height = sum(i.z for i in vertices) / len(vertices)
    imin, imax, jmin, jmax, _, _ = polygon.extreme
    xlen_spc = (imax - imin) / spacing
    ylen_spc = (jmax - jmin) / spacing
    xstart = ((xlen_spc - int(xlen_spc) + 1)) * spacing / 2
    ystart = ((ylen_spc - int(ylen_spc) + 1)) * spacing / 2
    x0 = [x + xstart for x in frange_inc(imin, imax, spacing)]
    y0 = [x + ystart for x in frange_inc(jmin, jmax, spacing)]
    grid_dir = polygon.normal.reverse()
    grid_hgt = geom.Vector(0, 0, plane_height) + grid_dir.scale(height)
    raw_pts = [
        geom.Vector(round(i, 3), round(j, 3), round(grid_hgt.z, 3))
        for i in x0
        for j in y0
    ]
    scale_factor = 1 - 0.3 / (imax - imin)  # scale boundary down .3 meter
    _polygon = polygon.scale(
        geom.Vector(scale_factor, scale_factor, 0), polygon.centroid
    )
    _vertices = _polygon.vertices
    if polygon.normal == geom.Vector(0, 0, 1):
        # pt_incls = pt_inclusion(_vertices)
        _grid = [p for p in raw_pts if pt_inclusion(p, _vertices) > 0]
    else:
        # pt_incls = pt_inclusion(_vertices[::-1])
        _grid = [p for p in raw_pts if pt_inclusion(p, _vertices[::-1]) > 0]
    # _grid = [p for p in raw_pts if pt_incls.test_inside(p) > 0]
    grid = [p.to_list() + grid_dir.to_list() for p in _grid]
    return grid


def material_lib() -> Dict[str, Any]:
    """Generate a list of generic material primitives."""
    tmis = tmit2tmis(0.6)
    return {
        "neutral_lambertian_0.2": neutral_plastic_prim(
            "void", "neutral_lambertian_0.2", 0.2, 0, 0
        ),
        "neutral_lambertian_0.5": neutral_plastic_prim(
            "void", "neutral_lambertian_0.5", 0.5, 0, 0
        ),
        "neutral_lambertian_0.7": neutral_plastic_prim(
            "void", "neutral_lambertian_0.7", 0.7, 0, 0
        ),
        "glass_60": glass_prim("void", "glass_60", tmis, tmis, tmis),
    }


def gen_blinds(depth, width, height, spacing, angle, curve, movedown) -> str:
    """Generate genblinds command for genBSDF."""
    nslats = int(round(height / spacing, 0))
    slat_cmd = "!genblinds blindmaterial blinds "
    slat_cmd += f"{depth} {width} {height} {nslats} {angle} {curve}"
    slat_cmd += "| xform -rz -90 -rx -90 -t "
    slat_cmd += f"{-width/2} {-height/2} {-movedown}\n"
    return slat_cmd


def get_glazing_primitive(panes: List[PaneRGB]) -> Primitive:
    """Generate a BRTDfunc to represent a glazing system."""
    if len(panes) > 2:
        raise ValueError("Only double pane supported")
    name = "+".join([pane.measured_data.name for pane in panes])
    if len(panes) == 1:
        str_arg = [
            "10",
            "sr_clear_r",
            "sr_clear_g",
            "sr_clear_b",
            "st_clear_r",
            "st_clear_g",
            "st_clear_b",
            "0",
            "0",
            "0",
            "glaze1.cal",
        ]
        coated_real = 1 if panes[0].measured_data.coated_side == "front" else -1
        real_arg = [
            19,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            coated_real,
            *[round(i, 3) for i in panes[0].glass_rgb],
            *[round(i, 3) for i in panes[0].coated_rgb],
            *[round(i, 3) for i in panes[0].trans_rgb],
        ]
    else:
        s12t_r, s12t_g, s12t_b = panes[0].trans_rgb
        s34t_r, s34t_g, s34t_b = panes[1].trans_rgb
        if panes[0].measured_data.coated_side == "back":
            s2r_r, s2r_g, s2r_b = panes[0].coated_rgb
            s1r_r, s1r_g, s1r_b = panes[0].glass_rgb
        else:  # front or neither side coated
            s2r_r, s2r_g, s2r_b = panes[0].glass_rgb
            s1r_r, s1r_g, s1r_b = panes[0].coated_rgb
        if panes[1].measured_data.coated_side == "back":
            s4r_r, s4r_g, s4r_b = panes[1].coated_rgb
            s3r_r, s3r_g, s3r_b = panes[1].glass_rgb
        else:  # front or neither side coated
            s4r_r, s4r_g, s4r_b = panes[1].glass_rgb
            s3r_r, s3r_g, s3r_b = panes[1].coated_rgb
        str_arg = [
            "10",
            (
                f"if(Rdot,cr(fr({s4r_r:.3f}),ft({s34t_r:.3f}),fr({s2r_r:.3f})),"
                f"cr(fr({s1r_r:.3f}),ft({s12t_r:.3f}),fr({s3r_r:.3f})))"
            ),
            (
                f"if(Rdot,cr(fr({s4r_g:.3f}),ft({s34t_g:.3f}),fr({s2r_g:.3f})),"
                f"cr(fr({s1r_g:.3f}),ft({s12t_g:.3f}),fr({s3r_g:.3f})))"
            ),
            (
                f"if(Rdot,cr(fr({s4r_b:.3f}),ft({s34t_b:.3f}),fr({s2r_b:.3f})),"
                f"cr(fr({s1r_b:.3f}),ft({s12t_b:.3f}),fr({s3r_b:.3f})))"
            ),
            f"ft({s34t_r:.3f})*ft({s12t_r:.3f})",
            f"ft({s34t_g:.3f})*ft({s12t_g:.3f})",
            f"ft({s34t_b:.3f})*ft({s12t_b:.3f})",
            "0",
            "0",
            "0",
            "glaze2.cal",
        ]
        real_arg = [9, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    return Primitive("void", "BRTDfunc", name, str_arg, real_arg)


def get_flush_corner_rays_command(ray_cnt: int, xres: int) -> list:
    """Flush the corner rays from a fisheye view.

    Args:
        ray_cnt: ray count;
        xres: resolution of the square image;

    Returns:
        Command to generate cropped rays

    """
    cmd = [
        "rcalc",
        "-if6",
        "-of",
        "-e",
        f"DIM:{xres};CNT:{ray_cnt}",
        "-e",
        "pn=(recno-1)/CNT+.5",
        "-e",
        "frac(x):x-floor(x)",
        "-e",
        "xpos=frac(pn/DIM);ypos=pn/(DIM*DIM)",
        "-e",
        "incir=if(.25-(xpos-.5)*(xpos-.5)-(ypos-.5)*(ypos-.5),1,0)",
        "-e",
        "$1=$1;$2=$2;$3=$3;$4=$4*incir;$5=$5*incir;$6=$6*incir",
    ]
    return cmd


def batch_process(
    commands: List[List[str]],
    inputs: Optional[List[bytes]] = None,
    opaths: Optional[List[Path]] = None,
    nproc: Optional[int] = None,
) -> None:
    """Run commands in batches.

    Use subprocess.Popen to run commands.

    Args:
        commands: commands as a list of strings.
        inputs: list of standard input to the commands.
        opaths: list of paths to write standard output to.
        nproc: number of commands to run in parallel at a time.
    Returns:
        None
    """
    nproc = 1 if nproc is None else nproc
    command_groups = [commands[i : i + nproc] for i in range(0, len(commands), nproc)]
    stdin_groups: List[List[Any]] = [[None] * len(ele) for ele in command_groups]
    if inputs:
        if len(inputs) != len(commands):
            raise ValueError("Number of stdins not equal number of commands.")
        stdin_groups = [inputs[i : i + nproc] for i in range(0, len(inputs), nproc)]
    if opaths:
        if len(opaths) != len(commands):
            raise ValueError("Number of opaths not equal number of commands.")
        opath_groups = [opaths[i : i + nproc] for i in range(0, len(opaths), nproc)]
        for igrp, cgrp, ogrp in zip(stdin_groups, command_groups, opath_groups):
            processes = []
            fds = []
            for command, opath in zip(cgrp, ogrp):
                fd = open(opath, "wb")
                processes.append(sp.Popen(command, stdin=sp.PIPE, stdout=fd))
                fds.append(fd)
            for proc, sin in zip(processes, igrp):
                if (proc.stdin is not None) and (sin is not None):
                    proc.stdin.write(sin)
            for p, f in zip(processes, fds):
                p.wait()
                f.close()
    else:
        for igrp, cgrp in zip(stdin_groups, command_groups):
            processes = [sp.Popen(command, stdin=sp.PIPE) for command in cgrp]
            for proc, sin in zip(processes, igrp):
                if (proc.stdin is not None) and (sin is not None):
                    proc.stdin.write(sin)
            for proc in processes:
                proc.wait()


def run_write(command, out, stdin=None) -> None:
    """Run command and write stdout to file."""
    with open(out, "wb") as wtr:
        sp.run(command, check=True, input=stdin, stdout=wtr)



"""
Routines for generating sky models
"""

import logging
import math
import os
from pathlib import Path
import subprocess as sp
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from frads import geom
from frads import parsers
from frads.types import WeaMetaData
from frads.types import WeaData
from frads import utils

logger: logging.Logger = logging.getLogger("frads.sky")

# Solar disk solid angle (sr)
SOLAR_SA = 6.7967e-5


def basis_glow(sky_basis: str) -> str:
    """
    Generate a set of regular sky and ground glow primitives string.

    Args:
        sky_basis(str): sky sampling basis, e.g. r1, r4
    Returns:
        ground and sky glow string, usually used for rfluxmtx calls.
    """
    grnd_str = grndglow()
    sky_str = skyglow(sky_basis)
    return grnd_str + sky_str


def skyglow(basis: str, upvect: str = "+Y") -> str:
    """
    Generate a set of skyglow string

    Args:
        basis(str): e.g., r1, r2, r4
        upvect(str): Optional, default=+Y
    Returns:
        A set of sky glow primitive string
    """
    sky_string = f"#@rfluxmtx u={upvect} h={basis}\n\n"
    sky_string += "void glow skyglow\n"
    sky_string += "0\n0\n4 1 1 1 0\n\n"
    sky_string += "skyglow source sky\n"
    sky_string += "0\n0\n4 0 0 1 180\n"
    return sky_string


def grndglow(basis: str = "u") -> str:
    """
    Generate a set of ground string
    Args:
        basis(str): Optional default=u
    Returns:
        A set of ground glow primitive string
    """
    ground_string = f"#@rfluxmtx h={basis}\n\n"
    ground_string += "void glow groundglow\n"
    ground_string += "0\n0\n4 1 1 1 0\n\n"
    ground_string += "groundglow source ground\n"
    ground_string += "0\n0\n4 0 0 -1 180\n\n"
    return ground_string


def gen_sun_source_full(mf: int) -> Tuple[str, str]:
    """
    Generate a full set of sun light sources according to Reinhart basis.

    Args:
        mf(int): multiplication factor, usually 1, 2, or 4.
    Returns:
        A tuple of full set of sun light and source primitive string
        and associated modifier string.
    """
    runlen = 144 * mf**2 + 3
    mod_str = os.linesep.join([f"sol{i}" for i in range(1, runlen)])
    dirs, _ = utils.calc_reinsrc_dir(mf)
    lines = []
    for i, d in enumerate(dirs):
        lines.append(
            f"void light sol{i} 0 0 3 1 1 1 sol{i} source sun "
            f"0 0 4 {d.x:.6g} {d.y:.6g} {d.z:.6g} 0.533"
        )
    return os.linesep.join(lines) + os.linesep, mod_str


def gen_sun_source_culled(
    mf,
    smx_path: Optional[Path] = None,
    window_normals: Optional[List[geom.Vector]] = None,
) -> Tuple[str, str, str]:
    """
    Generate a culled set of sun sources based on either window orientation
    and/or climate-based sky matrix. The reduced set of sun sources will
    significantly speed up the direct-sun matrix generation.

    Args:
        mf(int): multiplication factor, usually 1, 2, or 4.
        smx_path(str): Optional, sky matrix path, usually the output of gendaymtx
        window_normals(str): Optional, window normals
    Returns:
        A tuple of culled set of sun light and source primitive string,
        corresponding modifier strings, and the full set of modifier string.
    """
    runlen = 144 * mf**2 + 3
    dirs, _ = utils.calc_reinsrc_dir(mf)
    full_mod_str = os.linesep.join([f"sol{i}" for i in range(1, runlen)])
    win_norm = []
    if smx_path is not None:
        cmd1 = ["rmtxop", "-ff", "-c", ".3", ".6", ".1", "-t", str(smx_path)]
        cmd2 = ["getinfo", "-"]
        cmd3 = ["total", f"-if{runlen-1}", "-t,"]
        proc1 = sp.run(cmd1, check=True, stdout=sp.PIPE)
        proc2 = sp.run(cmd2, check=True, input=proc1.stdout, stdout=sp.PIPE)
        proc3 = sp.run(cmd3, check=True, input=proc2.stdout, stdout=sp.PIPE)
        dtot = [float(i) for i in proc3.stdout.split(b",")]
    else:
        dtot = [1] * runlen
    out_lines = []
    mod_str = []
    if window_normals is not None:
        win_norm = window_normals
        for i, d in enumerate(dirs):
            _mod = "sol" + str(i)
            v = 0
            if dtot[i] > 0:
                for norm in win_norm:
                    if norm * d < 0:
                        v = 1
                        mod_str.append(_mod)
                        break
            out_lines.append(
                f"void light sol{i} 0 0 3 {v} {v} {v} sol{i} source sun "
                f"0 0 4 {d.x:.6g} {d.y:.6g} {d.z:.6g} 0.533"
            )
    else:
        for i, d in enumerate(dirs):
            _mod = f"sol{i}"
            v = 0
            if dtot[i] > 0:
                v = 1
                mod_str.append(_mod)
            out_lines.append(
                f"void light sol{i} 0 0 3 {v} {v} {v} sol{i} source sun "
                f"0 0 4 {d.x:.6g} {d.y:.6g} {d.z:.6g} 0.533"
            )
    logger.debug(out_lines)
    logger.debug(mod_str)
    return os.linesep.join(out_lines), os.linesep.join(mod_str), full_mod_str


def gendaymtx(
    out: Union[str, Path],
    mf: int,
    data: Optional[Sequence[WeaData]] = None,
    meta: Optional[WeaMetaData] = None,
    wpath: Optional[Path] = None,
    direct: bool = False,
    solar: bool = False,
    onesun: bool = False,
    rotate: Optional[float] = None,
    binary: bool = False,
) -> List[str]:
    """
    Call gendaymtx to generate a sky/sun matrix
    and write results to out.  It takes either a .wea file path
    or wea data and metadata (defined in frads.types).
    If both are provided, .wea file path will be used.

    Args:
        out(str or pathlib.Path): outpath file path
        mf(int): multiplication factor
        data(Sequence[WeaData], optional): A sequence of WeaData.
        meta(WeaMetaData, optional): A instance of WeaMetaData object.
        wpath(Path, optional): .wea file path.
        direct(bool, optional): Whether to generate sun-only sky matrix.
        solar(bool, optional): Whether to generate sky matrix if solar spectrum.
        onesun(bool, optional): Whether to generate single sun matrix (five-phase).
        rotate(float, optional): rotate the sky counter-clock wise, looking down.
        binary(bool, optional): Whether to have outputs in single precision floats.
    Returns:
        cmd(List[str]): the gendaymtx command called.
    Raises:
        ValueError: An error occurs if neither a .wea path nor wea data is provided.
    """
    stdin = None
    cmd = ["gendaymtx", "-m", str(mf)]
    if binary:
        cmd.append("-of")
    if direct:
        cmd.append("-d")
    if onesun:
        cmd.extend(["-5", ".533"])
    if rotate is not None:
        cmd.extend(["-r", str(rotate)])
    if solar:
        cmd.append("-O1")
    if wpath is not None:
        cmd.append(str(wpath))
    elif (data is not None) and (meta is not None):
        wea_input = meta.wea_header() + "\n".join(map(str, data))
        stdin = wea_input.encode("utf-8")
    else:
        raise ValueError("Need to specify either .wea path or wea data.")
    with open(out, "wb") as wtr:
        logger.info("Calling gendaymtx with:\n%s", " ".join(cmd))
        sp.run(cmd, check=True, input=stdin, stdout=wtr)
    return cmd


def gen_perez_sky(row: WeaData, meta, grefl: float = 0.2, spect: str = "0", rotate=None) -> str:
    solar = False if spect == "0" else True
    gendaylit = gendaylit_cmd(
        str(row.time.month),
        str(row.time.day),
        str(row.time.hour + row.time.minute / 60 + row.time.second / 3600),
        str(meta.latitude),
        str(meta.longitude),
        str(meta.timezone),
        dir_norm_ir=str(row.dni),
        dif_hor_ir=str(row.dhi),
        solar=solar,
        grefl=grefl,
    )
    out = []
    rot = f"| xform -rz {rotate}" if rotate is not None else ""
    out.append(f"!{' '.join(gendaylit)}{rot} \n")
    out.append("skyfunc glow sglow 0 0 4 1 1 1 0\n")
    out.append("sglow source sky 0 0 4 0 0 1 180\n")
    out.append("sglow source ground 0 0 4 0 0 -1 180\n")
    return " ".join(out)


def gendaylit_cmd(
    month: str,
    day: str,
    hours: str,
    lat: str,
    lon: str,
    tzone: str,
    year: Optional[str] = None,
    grefl: Optional[float] = None,
    dir_norm_ir: Optional[str] = None,
    dif_hor_ir: Optional[str] = None,
    dir_hor_ir: Optional[str] = None,
    dir_norm_il: Optional[str] = None,
    dif_hor_il: Optional[str] = None,
    solar: bool = False,
) -> list:
    """Get a gendaylit command as a list."""
    cmd = ["gendaylit", month, day, hours]
    cmd += ["-a", lat, "-o", lon, "-m", tzone]
    if grefl is not None:
        cmd += ["-g", str(grefl)]
    if year is not None:
        cmd += ["-y", year]
    if None not in (dir_norm_ir, dif_hor_ir):
        cmd += ["-W", str(dir_norm_ir), str(dif_hor_ir)]
    if None not in (dir_hor_ir, dif_hor_ir):
        cmd += ["-G", str(dir_hor_ir), str(dif_hor_ir)]
    if None not in (dir_norm_il, dif_hor_il):
        cmd += ["-L", str(dir_norm_il), str(dif_hor_il)]
    if solar:
        cmd += ["-O", "1"]
    return cmd


def solar_angle(
    lat: float,
    lon: float,
    mer: float,
    month,
    day: int,
    hour,
) -> Tuple[float, float]:
    """
    Simplified translation from the Radiance sun.c and gensky.c code.
    """
    latitude_r = math.radians(lat)
    longitude_r = math.radians(lon)
    s_meridian = math.radians(mer)
    mo_da = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]

    julian_date = mo_da[month - 1] + day

    solar_decline = 0.4093 * math.sin((2 * math.pi / 365) * (julian_date - 81))

    solar_time = hour + (
        0.170 * math.sin((4 * math.pi / 373) * (julian_date - 80))
        - 0.129 * math.sin((2 * math.pi / 355) * (julian_date - 8))
        + (12 / math.pi) * (s_meridian - longitude_r)
    )
    altitude = math.asin(
        math.sin(latitude_r) * math.sin(solar_decline)
        - math.cos(latitude_r)
        * math.cos(solar_decline)
        * math.cos(solar_time * (math.pi / 12))
    )
    azimuth = -math.atan2(
        math.cos(solar_decline) * math.sin(solar_time * (math.pi / 12.0)),
        -math.cos(latitude_r) * math.sin(solar_time)
        - math.sin(latitude_r)
        * math.cos(solar_decline)
        * math.cos(solar_time * (math.pi / 12)),
    )

    return altitude, azimuth


def start_end_hour(data: Sequence[WeaData], sh: float, eh: float) -> Sequence[WeaData]:
    """Remove wea data entries outside of the
    start and end hour."""
    if sh == 0 and eh == 0:
        return data
    return [row for row in data if sh <= (row.time.hour + row.time.minute / 60) <= eh]


def check_sun_above_horizon(data, metadata):
    """Remove non-daylight hour entries."""

    def solar_altitude_check(row: WeaData):
        alt, _ = solar_angle(
            metadata.latitude,
            metadata.longitude,
            metadata.timezone,
            row.time.month,
            row.time.day,
            row.time.hour + row.time.minute / 60,
        )
        return alt > 0

    return [row for row in data if solar_altitude_check(row)]


def filter_data_with_zero_dni(data):
    """Filter out data entries with zero direct normal irradiance."""
    return [row for row in data if row.dni != 0]


def solar_minute(data: WeaData) -> int:
    # assuming never leap year
    mo_da = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334]
    jd = mo_da[data.time.month - 1] + data.time.day
    return 24 * 60 * (jd - 1) + int(data.time.hour * 60.0 + data.time.minute + 0.5)


def filter_data_by_direct_sun(
    data: Sequence[WeaData],
    meta: WeaMetaData,
    window_normal: Optional[Sequence[geom.Vector]] = None,
) -> List[WeaData]:
    """
    Remove wea data entries with zero solar luminance according to
    Perez All-Weather sky model. If window normal supplied,
    eliminate entries not seen by window. Window field of view
    is 176 deg with 2 deg tolerance on each side.

    Args:
        data: Sequence[WeaData],
        meta: WeaMetaData,
        window_normal: Optional[Sequence[geom.Vector]] = None,
    Returns:
        data(List[WeaData]):
    """
    wea_input = meta.wea_header() + "\n".join(map(str, data))
    proc = sp.run(
        ["gendaymtx", "-u", "-D", "-"],
        check=True,
        input=wea_input,
        encoding="ascii",
        stderr=sp.PIPE,
        stdout=sp.PIPE,
    )
    prims = parsers.parse_primitive(proc.stdout.splitlines())
    light_prims = [prim for prim in prims if prim.ptype == "light"]
    keep_minutes = []
    if window_normal is not None:
        source_prims = [prim for prim in prims if prim.ptype == "source"]
        for lpr, spr in zip(light_prims, source_prims):
            if lpr.real_arg[1] > 0:
                sdir = geom.Vector(*spr.real_arg[1:4])
                for normal in window_normal:
                    if normal * sdir < -0.035:  # 2deg tolerance
                        keep_minutes.append(int(spr.modifier.lstrip("solar")))
                        break
    else:
        for lpr in light_prims:
            if lpr.real_arg[1] > 0:
                keep_minutes.append(int(lpr.identifier.lstrip("solar")))
    # inminutes = [solar_minute(d) for d in data]
    inminutes = []
    extra_day = 0
    for d in data:
        inm = solar_minute(d)
        if d.time.month == 2 and d.time.day == 29:
            extra_day = 1440
        inm += extra_day
        inminutes.append(inm)
    new_dataline = [data for data, minu in zip(data, inminutes) if minu in keep_minutes]
    return new_dataline


def filter_wea(
    wea_data: Sequence[WeaData],
    meta_data: WeaMetaData,
    start_hour: Optional[float] = None,
    end_hour: Optional[float] = None,
    daylight_hours_only: bool = False,
    remove_zero: bool = False,
    window_normals: Optional[List[geom.Vector]] = None,
) -> Tuple[Sequence[WeaData], List[Any]]:
    """
    Obtain and prepare weather file data.

    Args:
        wea_data(List[WeaData]): A list of WeaData.
        meta_data(WeaMetaData): A instance of WeaMetaData object.
        start_hour(float, optional): Filter out wea data before this hour.
        end_hour(float, optional): Filter out wea data after this hour.
        daylight_hours_only(bool, optional): Filter out wea data below horizon.
        remove_zero(bool, optional): Filter out wea data with zero DNI.
        window_normals(List[geom.Vector], optional): Filter out wea data with direct
            sun not seen by these window normals.
    Returns:
        wea_data(List[WeaData]): Filterd list of wea data
        datetime_stamps(list): Remaining datetime stamps
    """
    logger.info("Filtering wea data, starting with %d rows", len(wea_data))
    if (start_hour is not None) and (end_hour is not None):
        wea_data = start_end_hour(wea_data, start_hour, end_hour)
        logger.info(
            "Filtering out hours outside of %f and %f: %d rows remaining",
            start_hour,
            end_hour,
            len(wea_data),
        )
    if daylight_hours_only:
        wea_data = check_sun_above_horizon(wea_data, meta_data)
        logger.info("Filtering by daylight hours: %d rows remaining", len(wea_data))
    if remove_zero:
        wea_data = filter_data_with_zero_dni(wea_data)
    if window_normals is not None:
        wea_data = filter_data_by_direct_sun(
            wea_data, meta_data, window_normal=window_normals
        )
        logger.info(
            "Filtering zero DNI hours and suns not seen by window: %d rows remaining",
            len(wea_data),
        )
    datetime_stamps = [row.dt_str() for row in wea_data]
    if len(wea_data) == 0:
        logger.warning("Empty wea file")
    return wea_data, datetime_stamps



# frads/types.py
"""
This module contains all data types used across frads.
The exceptions are the Vector and Polygon class in the geom.py module.

"""

import datetime
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from frads.geom import Polygon
from frads.geom import Vector


class Primitive(NamedTuple):
    """Radiance Primitive.

    Attributes one-to-one mapped from Radiance.

    Attributes:
        modifier: modifier, which primitive modifies this one
        ptype: primitive type
        identifier: identifier, name of this primitive
        str_arg: string argument
        real_arg: real argument
        int_arg: integer argument, not used in Radiance (default="0")
    """

    modifier: str
    ptype: str
    identifier: str
    str_arg: Sequence[str]
    real_arg: Sequence[Union[int, float]]
    int_arg: str = "0"

    def __repr__(self) -> str:
        output = (
            f"Primitive({self.modifier}, {self.ptype}, "
            f"{self.identifier}, {self.str_arg}, "
            f"{self.int_arg}, {self.real_arg})"
        )
        return output

    def __str__(self) -> str:
        output = (
            f"{self.modifier} {self.ptype} {self.identifier}\n"
            f"{int(self.str_arg[0])} {' '.join(self.str_arg[1:])}\n"
            f"{self.int_arg}\n"
            f"{int(self.real_arg[0])} "
            f"{' '.join(map(str, self.real_arg[1:]))}\n"
        )
        return output


@dataclass
class Alias:
    modifier: str
    name_to: str
    name_from: str

    def __repr__(self) -> str:
        output = f"Alias({self.modifier}, {self.name_to}, {self.name_from})"
        return output

    def __str__(self) -> str:
        output = f"{self.modifier} alias {self.name_to} {self.name_from}"
        return output


@dataclass(frozen=True)
class Sender:
    """Sender object for matrix generation.

    Attributes:
        form: types of sender, {surface(s)|view(v)|points(p)}
        sender: the sender string
        xres: sender x dimension
        yres: sender y dimension
    """

    form: str
    sender: bytes
    xres: Optional[int]
    yres: Optional[int]


@dataclass
class Receiver:
    """Receiver object for matrix generation.

    Attributes:
        receiver: receiver string which can be appended to one another
        basis: receiver basis, usually kf, r4, r6;
        modifier: modifiers to the receiver objects;
    """

    receiver: str
    basis: str
    modifier: str = ""

    def __add__(self, other) -> "Receiver":
        return Receiver(
            self.receiver + "\n" + other.receiver, self.basis, self.modifier
        )


@dataclass
class ScatteringData:
    """Scattering data object.

    Attributes:
        sdata: scattering data in nested lists.
        ncolumn: number of columns
        nrow: number of rows
    """

    sdata: List[float]
    ncolumn: int
    nrow: int

    def __str__(self) -> str:
        out = "#?RADIANCE\nNCOMP=3\n"
        out += f"NROWS={self.nrow}\nNCOLS={self.ncolumn}\n"
        out += "FORMAT=ascii\n\n"
        for col in range(self.ncolumn):
            for row in range(self.nrow):
                val = self.sdata[row + col * self.ncolumn]
                string = "\t".join([f"{val:7.5f}"] * 3)
                out += string + "\t"
            out += "\n"
        return out


@dataclass
class BSDFData:
    """BSDF data object.

    Attributes:
        bsdf: BSDF data.
        ncolumn: number of columns
        nrow: number of rows
    """

    bsdf: List[float]
    ncolumn: int
    nrow: int


@dataclass(frozen=True)
class RadMatrix:
    """Radiance matrix object.

    Attributes:
        tf: front-side transmission
        tb: back-side transmission
    """

    tf: ScatteringData
    tb: ScatteringData


class PaneProperty(NamedTuple):
    """Window pane property object.

    Attributes:
        name str: material name.
        thickness float: pane thickness.
        gtype str: material type.
        coated_side str: coated side, front or back.
        wavelength: Wavelength data.
        transmittance: Transmittance data.
        reflectance_front: Reflectance front data.
        reflectance_back: Reflectance back data.
    """

    name: str
    thickness: float
    gtype: str
    coated_side: str
    wavelength: List[float]
    transmittance: List[float]
    reflectance_front: List[float]
    reflectance_back: List[float]

    def get_tf_str(self) -> str:
        wavelength_tf = tuple(zip(self.wavelength, self.transmittance))
        return "\n".join([" ".join(map(str, pair)) for pair in wavelength_tf])

    def get_rf_str(self) -> str:
        wavelength_rf = tuple(zip(self.wavelength, self.reflectance_front))
        return "\n".join([" ".join(map(str, pair)) for pair in wavelength_rf])

    def get_rb_str(self) -> str:
        wavelength_rb = tuple(zip(self.wavelength, self.reflectance_back))
        return "\n".join([" ".join(map(str, pair)) for pair in wavelength_rb])


class PaneRGB(NamedTuple):
    """Pane color data object.

    Attributes:
        measured_data: measured data as a PaneProperty object.
        coated_rgb: Coated side RGB.
        glass_rgb: Non-coated side RGB.
        trans_rgb: Transmittance RGB.
    """

    measured_data: PaneProperty
    coated_rgb: Iterable[float]
    glass_rgb: Iterable[float]
    trans_rgb: Iterable[float]


class WeaMetaData(NamedTuple):
    """Weather related meta data object.

    Attributes:
        city: City.
        country: Country.
        latitude: Latitude.
        longitude: Longitude.
        timezone: Timezone as standard meridian.
        elevation: Site elevation (m).
    """

    city: str
    country: str
    latitude: float
    longitude: float
    timezone: int
    elevation: float

    def wea_header(self) -> str:
        """Return a .wea format header."""
        return (
            f"place {self.city}_{self.country}\n"
            f"latitude {self.latitude}\n"
            f"longitude {self.longitude}\n"
            f"time_zone {self.timezone}\n"
            f"site_elevation {self.elevation}\n"
            "weather_data_file_units 1\n"
        )


class WeaData(NamedTuple):
    """Weather related data object.

    Attributes:
        month: Month.
        day: Day.
        hour: Hour.
        minute: Minutes.
        second: Seconds.
        hours: Times with minutes as fraction.
        dni: Direct normal irradiance (W/m2).
        dhi: Diffuse horizontal irradiance (W/m2).
        aod: Aeroal Optical Depth (default = 0).
        cc: Cloud cover (default = 0).
        year: default = 2000.
    """

    time: datetime.datetime
    dni: float
    dhi: float
    aod: float = 0
    cc: float = 0

    def __str__(self) -> str:
        return f"{self.time.month} {self.time.day} {self.time.hour+self.time.minute/60} {self.dni} {self.dhi}"

    def dt_str(self) -> str:
        return f"{self.time.month:02d}{self.time.day:02d}_{self.time.hour:02d}{self.time.minute:02d}"


class MradModel(NamedTuple):
    """Mrad model object.
    Attributes:
        name: Model name
        material_path: Material path
        window_groups: Window primitives grouped by files
        window_normals: Window normals
        sender_grid: Grid ray samples mapped to grid surface name.
        sender_view: View ray samples mapped to view name.
        views: Mapping from View name to view properties.
        receiver_sky: Sky as the receiver object.
        bsdf_xml: Mapping from window groupd name to BSDF file path.
        cfs_paths: The list of files used for direct-sun coefficient calculations.
        ncp_shades: The list of non-coplanar shading files.
        black_env_path: Blackened environment file path.

    """

    name: str
    material_path: Path
    window_groups: Dict[str, List[Primitive]]
    window_normals: List[Vector]
    sender_grid: dict
    sender_view: dict
    views: dict
    bsdf_xml: dict
    cfs_paths: list
    ncp_shades: dict
    black_env_path: Path


@dataclass
class MradPath:
    """
    This dataclass object holds all the paths during a mrad run.
    All attributes are initiated with default_factory set to the attribute's type,
    which means this object can be instantiated without any arguments and
    add define its attributes later.

    Attributes:
        pvmx: Point view matrix paths mapped to grid name.
        pvmxd: Direct only point view matrix paths mapped to grid name.
        pdsmx: Point daylight coefficient matrix file paths mapped to grid name.
        pcdsmx: Point direct-sun coefficient matrix file paths mapped to grid name.
        vvmx: View view matrix paths mapped to view name.
        vvmxd: Direct only view view matrix paths mapped to view name.
        vdsmx: View daylight coefficient matrix file paths mapped to grid name.
        vcdsmx: View direct-sun coefficient matrix file paths mapped to view name.
        vcdfmx: View direct-sun coefficient(f) matrix file paths mapped to view name.
        vcdrmx: View direct-sun coefficient(r) matrix file paths mapped to view name.
        vmap: View matrix material map mapped to view name.
        cdmap: Direct-sun matrix material map mapped to view name.
        dmx: Daylight matrix file paths mapped to window name.
        dmxd: Direct daylight matrix mapped to window name.
        smxd: Sun-only (3-4 sun patches) sky matrix file path.
        smx: sky matrix file path.
        smx_sun: Sun-only (one sun patch) sky matrix file path for illuminance.
        smx_sun_img: Sun-only (one sun pathc) sky matrix file path for rendering.
    """

    pvmx: Dict[str, Path] = field(default_factory=dict)
    vvmx: Dict[str, Path] = field(default_factory=dict)
    dmx: Dict[str, Path] = field(default_factory=dict)
    pdsmx: Dict[str, Path] = field(default_factory=dict)
    vdsmx: Dict[str, Path] = field(default_factory=dict)
    pcdsmx: Dict[str, Path] = field(default_factory=dict)
    vcdsmx: Dict[str, Path] = field(default_factory=dict)
    vcdfmx: Dict[str, Path] = field(default_factory=dict)
    vcdrmx: Dict[str, Path] = field(default_factory=dict)
    vmap: Dict[str, Path] = field(default_factory=dict)
    cdmap: Dict[str, Path] = field(default_factory=dict)
    smxd: Path = field(default_factory=Path)
    pvmxd: Dict[str, Path] = field(default_factory=dict)
    vvmxd: Dict[str, Path] = field(default_factory=dict)
    dmxd: Dict[str, Path] = field(default_factory=dict)
    smx: Path = field(default_factory=Path)
    smx_sun: Path = field(default_factory=Path)
    smx_sun_img: Path = field(default_factory=Path)


@dataclass
class NcpModel:
    """Non-coplanar data model."""

    windows: Sequence[Primitive]
    ports: Sequence[Primitive]
    env: Iterable[Path]
    sbasis: str
    rbasis: str


@dataclass
class EPlusWindowGas:
    """EnergyPlus Window Gas material data container."""

    name: str
    thickness: float
    type: list
    percentage: list
    primitive: str = ""


@dataclass
class EPlusOpaqueMaterial:
    """EnergyPlus Opaque material data container."""

    name: str
    roughness: str
    solar_absorptance: float
    visible_absorptance: float
    visible_reflectance: float
    primitive: Primitive
    thickness: float = 0.0


@dataclass
class EPlusWindowMaterial:
    """EnergyPlus regular window material data container."""

    name: str
    visible_transmittance: float
    primitive: Primitive


@dataclass
class EPlusConstruction:
    """EnergyPlus construction data container."""

    name: str
    type: str
    layers: list


@dataclass
class EPlusOpaqueSurface:
    """EnergyPlus opaque surface data container."""

    name: str
    type: str
    polygon: Polygon
    construction: str
    boundary: str
    sun_exposed: bool
    zone: str
    fenestrations: list


@dataclass
class EPlusFenestration:
    """EnergyPlus fenestration data container."""

    name: str
    type: str
    polygon: Polygon
    construction: EPlusConstruction
    host: EPlusOpaqueSurface


@dataclass
class EPlusZone:
    """EnergyPlus zone data container."""

    name: str
    wall: Dict[str, EPlusOpaqueSurface]
    ceiling: Dict[str, EPlusOpaqueSurface]
    roof: Dict[str, EPlusOpaqueSurface]
    floor: Dict[str, EPlusOpaqueSurface]
    window: Dict[str, EPlusFenestration]


class ViewType:
    perspective = "v"
    parallel = "l"
    cylindrical = "c"
    hemispherical_fisheye = "h"
    angular_fisheye = "a"
    planisphere_fishsye = "s"


@dataclass
class View:
    """View data object."""

    position: Vector
    direction: Vector
    up: Vector = Vector(0, 0, 1)
    vtype: str = "v"
    vert: float = 45
    hori: float = 45
    aft: float = 0
    fore: float = 0
    shift: float = 0
    lift: float = 0
    xres: int = 256
    yres: int = 256

    def __str__(self) -> str:
        return (
            f"-vt{self.vtype} "
            f"-vp {self.position} -vd {self.direction} "
            f"-vu {self.up} "
            f"-vv {self.vert} -vh {self.hori}"
            f"-x {self.xres} -y {self.yres}"
        )

    def args(self) -> list:
        return [
            f"-vt{self.vtype}",
            "-vp",
            *map(str, self.position.to_list()),
            "-vd",
            *map(str, self.direction.to_list()),
            "-vu",
            *map(str, self.up.to_list()),
            "-vv",
            str(self.vert),
            "-vh",
            str(self.hori),
            "-vo",
            str(self.fore),
            "-va",
            str(self.aft),
            "-vs",
            str(self.shift),
            "-vl",
            str(self.lift),
            "-x",
            str(self.xres),
            "-y",
            str(self.yres),
        ]


class ColorPrimaries(NamedTuple):
    xr: float
    yr: float
    xg: float
    yg: float
    xb: float
    yb: float
    xw: float
    yw: float


class IntArg:
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, type=None) -> object:
        return obj.__dict__.get(self.name)

    def __set__(self, obj, value) -> None:
        if not isinstance(value, int):
            raise ValueError("ab has to be an integer")
        self.value = value
        obj.__dict__[self.name] = value


class FloatArg:
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, type=None) -> object:
        return obj.__dict__.get(self.name)

    def __set__(self, obj, value) -> None:
        if not isinstance(value, float):
            raise ValueError("Value has to be a float")
        self.value = value
        obj.__dict__[self.name] = value


class TupleArg:
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, obj, type=None) -> object:
        return obj.__dict__.get(self.name)

    def __set__(self, obj, value) -> None:
        if not isinstance(value, tuple):
            raise ValueError("Value has to be a tuple")
        if not all(isinstance(i, (int, float)) for i in value):
            raise ValueError("Value inside has to be a integer or float")
        self.value = value
        obj.__dict__[self.name] = value


class Options:
    aa: float = FloatArg()
    ab: int = IntArg()
    ad: int = IntArg()
    ar: int = IntArg()
    as_: int = IntArg()
    av: Tuple[float, float, float] = TupleArg()
    aw: int = IntArg()
    dc: float = FloatArg()
    dj: float = FloatArg()
    dr: int = IntArg()
    dp: int = IntArg()
    ds: float = FloatArg()
    dt: float = FloatArg()
    lr: int = IntArg()
    lw: float = FloatArg()
    ms: float = FloatArg()
    pa: float = FloatArg()
    pj: float = FloatArg()
    ps: int = IntArg()
    pt: float = FloatArg()
    ss: float = IntArg()
    st: float = FloatArg()
    x: int = IntArg()
    y: int = IntArg()
    i: bool = False
    I: bool = False

    def args(self):
        arg_list: List[str] = []
        for k, v in self.__dict__.items():
            if k in Options.__annotations__:
                if isinstance(v, bool):
                    sign = "+" if v else "-"
                    arg_list.append(f"-{k}{sign}")
                elif isinstance(v, (int, float)):
                    arg_list.append("-" + k)
                    arg_list.append(str(v))
                elif isinstance(v, tuple):
                    arg_list.append("-" + k)
                    arg_list.extend(map(str, v))
        return arg_list



"""
This module contains definitions of Vector and Polygon objects
and other geometry related routines.
"""

from __future__ import annotations

from dataclasses import dataclass
import math
from typing import Union
from typing import List
from typing import Tuple
from typing import Sequence
from typing import Iterable


@dataclass(frozen=True)
class Vector:
    """3D vector class.

    Attributes:
        x: x coordinate
        y: y coordinate
        z: z coordinate
    """

    x: float = 0
    y: float = 0
    z: float = 0

    def __str__(self) -> str:
        """Class string representation.

        Returns:
            The string representation of the vector(str)

        """
        return f"{self.x:02f} {self.y:02f} {self.z:02f}"

    def __add__(self, other) -> Vector:
        """Add the two vectors.
        Args:
            other(Vector): vector to add

        Returns:
            The added vector(Vector)

        """
        return Vector(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: Vector) -> Vector:
        return Vector(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: Vector) -> float:
        """Return the dot produce between two vectors."""
        return self.x * other.x + self.y * other.y + self.z * other.z

    @property
    def length(self) -> float:
        """Get vector distance from origin."""
        return math.sqrt(self.x**2 + self.y**2 + self.z**2)

    def cross(self, other: Vector) -> Vector:
        """Return the cross product of the two vectors.

        Args:
            other: the vector to take a cross product

        Returns:
            The resulting vector

        """
        return Vector(
            self.y * other.z - self.z * other.y,
            self.z * other.x - self.x * other.z,
            self.x * other.y - self.y * other.x,
        )

    def distance_from(self, other: Vector) -> float:
        """Calculate the distance between two points."""
        dx = math.fabs(self.x - other.x)
        dy = math.fabs(self.y - other.y)
        dz = math.fabs(self.z - other.z)
        return math.sqrt(dx**2 + dy**2 + dz**2)

    def normalize(self) -> Vector:
        """Return the unit vector."""
        magnitude = math.sqrt(self.x**2 + self.y**2 + self.z**2)
        return Vector(self.x / magnitude, self.y / magnitude, self.z / magnitude)

    def reverse(self) -> Vector:
        """Return the reversed vector."""
        return Vector(self.x * -1, self.y * -1, self.z * -1)

    def scale(self, factor: float) -> Vector:
        """Scale the vector by a scalar."""
        return Vector(self.x * factor, self.y * factor, self.z * factor)

    def angle_from(self, other: "Vector") -> float:
        """."""
        dot_prod = self * other
        angle = math.acos(dot_prod / (self.length * other.length))
        return angle

    def rotate_3d(self, vector: Vector, theta: float) -> Vector:
        """Rotate the point around the vector theta radians.

        Args:
            vector: rotation axis
            theta: rotation radians
        Returns:
            the rotated point

        """
        cosa = math.cos(theta)
        sina = math.sin(theta)

        row1 = [
            (vector.x * vector.x) + ((1 - (vector.x * vector.x)) * cosa),
            (vector.x * vector.y * (1 - cosa)) - (vector.z * sina),
            (vector.x * vector.z * (1 - cosa)) + (vector.y * sina),
            0.0,
        ]
        row2 = [
            (vector.x * vector.y * (1 - cosa)) + (vector.z * sina),
            (vector.y * vector.y) + ((1 - (vector.y * vector.y)) * cosa),
            (vector.y * vector.z * (1 - cosa)) - (vector.x * sina),
            0.0,
        ]
        row3 = [
            (vector.x * vector.z * (1.0 - cosa)) - (vector.y * sina),
            (vector.y * vector.z * (1.0 - cosa)) + (vector.x * sina),
            (vector.z * vector.z) + ((1.0 - (vector.z * vector.z)) * cosa),
            0.0,
        ]

        row4 = [0.0] * 4

        rx = (self.x * row1[0]) + (self.y * row2[0]) + (self.z * row3[0]) + row4[0]
        ry = (self.x * row1[1]) + (self.y * row2[1]) + (self.z * row3[1]) + row4[1]
        rz = (self.x * row1[2]) + (self.y * row2[2]) + (self.z * row3[2]) + row4[2]

        return Vector(rx, ry, rz)

    def to_sphr(self):
        """Convert cartesian to spherical coordinates."""
        r = self.distance_from(Vector())
        theta = math.atan(self.y / self.x)
        phi = math.acos(self.z / r)
        return theta, phi, r

    def coplanar(self, other1, other2) -> bool:
        """Test if the vector is coplanar with the other two vectors."""
        triple_prod = self * other1.cross(other2)
        return triple_prod == 0

    def to_list(self) -> List[float]:
        """Return a list containing the coordinates."""
        return [self.x, self.y, self.z]

    def to_tuple(self) -> Tuple[float, ...]:
        """Return a tuple containing the coordinates."""
        return self.x, self.y, self.z

    @classmethod
    def spherical(
        cls,
        theta: float,
        phi: float,
        r: float,
    ) -> Vector:
        """Construct a vector using spherical coordinates."""
        xcoord = math.sin(theta) * math.cos(phi) * r
        ycoord = math.sin(theta) * math.sin(phi) * r
        zcoord = math.cos(theta) * r
        return cls(xcoord, ycoord, zcoord)


@dataclass(frozen=True)
class Polygon:
    """3D polygon class

    Attributes:
        vertices(List[Vector]): list of vertices of the polygon.
    """

    vertices: Iterable[Vector]

    def __post_init__(self) -> None:
        """."""
        if len(self.vertices) < 3:
            raise ValueError("Need more than 2 vertices to make a polygon.")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Polygon):
            return NotImplemented
        if len(other.vertices) != len(self.vertices):
            return False
        for vert in self.vertices:
            if vert not in other.vertices:
                return False
        return True

    def __add__(self, other: Polygon) -> Polygon:
        """Merge two polygons."""
        sp, index = self.shared_pts(other)
        if sp != 2:
            raise ValueError("Trying to merge two polygons without shared sides")
        t1 = self.vertices[index[0] :] + self.vertices[: index[0]]
        oid = other.vertices.index(self.vertices[index[0]])
        t2 = other.vertices[oid:] + other.vertices[:oid]
        if t1[-1] == t2[1]:
            return Polygon(t1 + t2[2:])
        return Polygon(t2 + t1[2:])

    def __sub__(self, other: Polygon) -> Polygon:
        """Polygon subtraction.

        Args:
            other: subtract this polygon

        Returns:
            Clipped polygon (Polygon)

        """
        pt1 = self.vertices[0]
        # opposite = False if self.normal() == other.normal() else True
        opposite = False
        distances1 = [pt1.distance_from(i) for i in other.vertices]
        idx_min = distances1.index(min(distances1))
        new_other_vert = other.vertices[idx_min:] + other.vertices[:idx_min]
        results = [pt1]
        if opposite:
            results.extend(new_other_vert)
            results.append(other.vertices[idx_min])
        else:
            results.append(new_other_vert[0])
            results.extend(reversed(new_other_vert[1:]))
            results.append(new_other_vert[0])
        results.extend(self.vertices)
        return Polygon(tuple(results))

    @property
    def normal(self) -> Vector:
        """Calculate the polygon normal."""
        normal = Vector(0, 0, 0)
        for idx in range(len(self.vertices) - 2):
            normal += (self.vertices[idx + 1] - self.vertices[idx]).cross(
                self.vertices[idx + 2] - self.vertices[idx + 1]
            )
        return normal.normalize()

    @property
    def centroid(self) -> Vector:
        """Return the geometric center point."""
        return sum(self.vertices, Vector()).scale(1 / len(self.vertices))

    @property
    def area(self) -> float:
        """Calculate the area of the polygon."""
        total = Vector(0, 0, 0)
        for idx in range(1, len(self.vertices) - 1):
            total += (self.vertices[idx] - self.vertices[0]).cross(
                self.vertices[idx + 1] - self.vertices[0]
            )
        return abs(total * Vector(0.5, 0.5, 0.5))

    @property
    def extreme(self) -> Tuple[float, ...]:
        """."""
        xs = [v.x for v in self.vertices]
        ys = [v.y for v in self.vertices]
        zs = [v.z for v in self.vertices]
        return min(xs), max(xs), min(ys), max(ys), min(zs), max(zs)

    def flip(self) -> "Polygon":
        """Reverse the vertices order, thus reversing the normal."""
        return Polygon(self.vertices[::-1])
        # return Polygon([self.vertices[0]] + self.vertices[:0:-1])

    def scale(self, scale_vect, center) -> "Polygon":
        """Scale the polygon.

        Parameters:
            scale_vect (Vector): scale along x, y, z;
            center (Vector): center of scaling
        Return:
            Scaled polygon (Polygon)

        """
        new_vertices = []
        for vert in self.vertices:
            sx = center.x + (vert.x - center.x) * scale_vect.x
            sy = center.y + (vert.y - center.y) * scale_vect.y
            sz = center.z + (vert.z - center.z) * scale_vect.z
            new_vertices.append(Vector(sx, sy, sz))
        return Polygon(tuple(new_vertices))

    def extrude(self, vector: Vector) -> list:
        """Extrude the polygon.

        Args:
            vector (Vector): extrude along the vector;

        Returns:
            Polygon (list): a list of polygons;

        """
        polygons = [self]
        polygon2 = Polygon(([i + vector for i in self.vertices]))
        polygons.append(polygon2)
        for i in range(len(self.vertices) - 1):
            polygons.append(
                Polygon(
                    (
                        self.vertices[i],
                        polygon2.vertices[i],
                        polygon2.vertices[i + 1],
                        self.vertices[i + 1],
                    )
                )
            )
        polygons.append(
            Polygon(
                (
                    self.vertices[-1],
                    polygon2.vertices[-1],
                    polygon2.vertices[0],
                    self.vertices[0],
                )
            )
        )
        return polygons

    def shared_pts(self, other) -> Tuple[int, List[int]]:
        """Return the total number of share points between two polygons."""
        cnt = 0
        index = []
        for pid, val in enumerate(self.vertices):
            if val.to_list() in other.to_list():
                cnt += 1
                index.append(pid)
        return cnt, index

    def rotate(self, vector, angle) -> "Polygon":
        """."""
        ro_pts = [v.rotate_3d(vector, angle) for v in self.vertices]
        return Polygon(ro_pts)

    def move(self, vector) -> "Polygon":
        """Return the moved polygon along a vector."""
        mo_pts = [v + vector for v in self.vertices]
        return Polygon(mo_pts)

    def to_list(self):
        """Return a list of tuples."""
        return [p.to_tuple() for p in self.vertices]

    def to_real(self) -> Union[List[float], List[int]]:
        """Convert the vertices to real arg string format."""
        real_arg = [3 * len(self.vertices)]
        for vert in self.vertices:
            real_arg.append(vert.x)
            real_arg.append(vert.y)
            real_arg.append(vert.z)
        return real_arg

    @classmethod
    def rectangle3pts(cls, pt1, pt2, pt3) -> "Polygon":
        """."""
        vect21 = pt2 - pt1
        vect32 = pt3 - pt2
        err_msg = "Three points are not at a right angle"
        assert vect21.angle_from(vect32) == math.pi / 2, err_msg
        vect12 = pt1 - pt2
        pt4 = pt3 + vect12
        return cls((pt1, pt2, pt3, pt4))


def convexhull(points: List[Vector], normal: Vector) -> Polygon:
    """Convex hull on coplanar points.

    Parameters:
        points (list): list of Point;
        normal (Vector): plane's normal
    """

    def toleft(u, v, points):
        """."""
        vect2 = v - u
        pts = []
        for p in points:
            vect1 = p - u
            cross_prod = vect1.cross(vect2)
            cross_val = cross_prod * normal
            if cross_val < 0:
                pts.append(p)
        return pts

    def extend_pts(u, v, points):
        """."""
        if not points:
            return []

        vect2 = v - u
        w = min(points, key=lambda p: (p - u).cross(vect2) * normal)
        p1, p2 = toleft(w, v, points), toleft(u, w, points)
        return extend_pts(w, v, p1) + [w] + extend_pts(u, w, p2)

    u = min(points, key=lambda p: p.x)
    v = max(points, key=lambda p: p.x)
    if u == v:
        u = min(points, key=lambda p: p.y)
        v = max(points, key=lambda p: p.y)
    left, right = toleft(u, v, points), toleft(v, u, points)
    points = [v] + extend_pts(u, v, left) + [u] + extend_pts(v, u, right)
    points.append(points[0])
    points.append(points[1])
    to_remove = []
    for i, p in enumerate(points):
        if i < len(points) - 2:
            p12 = p.distance_from(points[i + 1])
            p23 = points[i + 1].distance_from(points[i + 2])
            p13 = p.distance_from(points[i + 2])
            if (p12 + p23) == p13:
                to_remove.append(points[i + 1])
    points.pop()
    points.pop()
    for r in to_remove:
        points.remove(r)
    return Polygon(points)


def polygon_center(*polygons):
    """Calculate the center from polygons."""
    vertices = [v for p in polygons for v in p.vertices]
    return sum(vertices, Vector()).scale(1 / len(vertices))


def get_polygon_limits(polygon_list: Sequence[Polygon], offset: float = 0.0):
    """Get the x,y,z limits from a list of polygons."""
    extreme_list = [p.extreme for p in polygon_list]
    lim = list(zip(*extreme_list))
    xmin = min(lim[0]) - offset
    xmax = max(lim[1]) + offset
    ymin = min(lim[2]) - offset
    ymax = max(lim[3]) + offset
    zmin = min(lim[4]) - offset
    zmax = max(lim[5]) + offset
    return xmin, xmax, ymin, ymax, zmin, zmax


def getbbox(polygons: Sequence[Polygon], offset: float = 0.0):
    """Get a bounding box for a list of polygons.

    Return a list of polygon that is the
    orthogonal bounding box of a list of polygon.

    Args:
        polygons: list of polygons
        offset: make the box smaller or bigger

    Returns:
        A list of polygon that is the bounding box.
    """
    xmin, xmax, ymin, ymax, zmin, zmax = get_polygon_limits(polygons, offset=offset)
    fp1 = Vector(xmin, ymin, zmin)
    fp2 = Vector(xmax, ymin, zmin)
    fp3 = Vector(xmax, ymax, zmin)
    fpg = Polygon.rectangle3pts(fp1, fp2, fp3)  # -Z
    cp1 = Vector(xmin, ymin, zmax)
    cp2 = Vector(xmax, ymin, zmax)
    cp3 = Vector(xmax, ymax, zmax)
    cpg = Polygon.rectangle3pts(cp3, cp2, cp1)  # +Z
    swpg = Polygon.rectangle3pts(cp2, fp2, fp1)  # -Y
    ewpg = Polygon.rectangle3pts(fp3, fp2, cp2)  # +X
    s2n_vec = Vector(0, ymax - ymin, 0)
    nwpg = Polygon([v + s2n_vec for v in swpg.vertices]).flip()  # +Y
    e2w_vec = Vector(xmax - xmin, 0, 0)
    wwpg = Polygon([v - e2w_vec for v in ewpg.vertices]).flip()  # -X

    return [fpg, cpg, ewpg, swpg, wwpg, nwpg]


def merge_polygon(polygons: Sequence[Polygon]) -> Polygon:
    """Merge polygons into a polygon using Convex Hull.

    Args:
        polygons: Polygons to be merged
    Returns:
        Merged polygon
    Raises:
        ValueError if window normals are not the same
    """
    normals = [p.normal for p in polygons]
    if len(set(normals)) > 1:
        raise ValueError("Windows not co-planar")
    points = [i for p in polygons for i in p.vertices]
    return convexhull(points, normals[0])



"""
Frads is an open-source library providing high-level
abstraction of Radiance matrix-based simulation workflows.

Frads automates setup of these simulations by providing
end users with an open-source, high-level abstraction of
the Radiance command-line workflow (Unix toolbox model),
helping to reduce the steep learning curve and associated
user errors. frads also provides the necessary infrastructure
needed for seamless integration of Radiance and other
modeling tools, such as EnergyPlus.

## Intended audience

1. Developers who are interested in incorporating multi-phase
matrix methods into their software and are seeking examples
and guidance; i.e., LBNL-suggested default parameters and settings; and,
2. Engineering firms, researchers, and students who are comfortable
working in the command-line or Python scripting environment and
tasked with a project that cannot be completed with existing tools.

## Why matrix-based methods?

Matrix algebraic methods reduce the time needed to perform accurate,
ray-tracing based, annual daylight simulations by several orders of
magnitude.

## Why frads?

A good deal of expertise is needed to set up the simulations
properly to achieve the desired level of accuracy.
frads provides users with tools (e.g., `mrad`) that automatically
determine which matrix-based method to use then sets the associated
simulation parameters, helping beginners learn the different matrix
methods by observing the tools’ behavior. The user is still required
to understand basic concepts underlying matrix-based simulation methods
(see [tutorials](https://www.radiance-online.org/learning/tutorials)).

Matrix-based methods also enable accurate, ray-tracing generated,
irradiance, illuminance, and luminance data to be available for
run-time data exchange and co-simulations. frads provides users with
tools that generate the appropriate Radiance-generated
data then interfaces with the “actuator” EMS module in EnergyPlus or
within the Spawn-of-EnergyPlus and Modelica co-simulation environment.
This enables end users to evaluate the performance of buildings with
manual- and automatically-controlled shading and daylighting systems
or other site and building features that can change parametrically
or on a time-step basis.
"""

import logging
import shutil
import subprocess as sp

from .matrix import (
    rfluxmtx,
    surface_as_sender,
    points_as_sender,
    view_as_sender,
    surface_as_receiver,
    sun_as_receiver,
    sky_as_receiver,
)

from .parsers import (
    parse_primitive,
    parse_epw,
    parse_wea,
    parse_polygon,
)

# from .raycall import oconv, rtrace, render

from .sky import basis_glow, gen_perez_sky

from .types import (
    Primitive,
    View,
    WeaData,
    WeaMetaData,
)

from .utils import gen_grid, unpack_primitives

__version__ = "0.2.10"

logger: logging.Logger = logging.getLogger(__name__)

# Check if Radiance is installed more or less
rad_progs = [
    "rfluxmtx",
    "total",
    "getinfo",
    "pcomb",
    "dctimestep",
    "rmtxop",
    "gendaymtx",
    "rtrace",
]

for prog in rad_progs:
    ppath = shutil.which(prog)
    if ppath is None:
        logger.info("%s not found; check Radiance installation", prog)

try:
    # Check Radiance version, need to be at least 5.X
    version_check: str = sp.run(
        ["rtrace", "-version"],
        check=True,
        stdout=sp.PIPE,
        encoding="ascii",
    ).stdout
    try:
        rad_version = float(version_check.split()[1][:3])
        if rad_version < 5.3:
            logger.critical("Please upgrade to Radiance version 5.3 or later")
    except ValueError:
        logger.critical(version_check)
except FileNotFoundError as err:
    logger.critical(err)

__all__ = [
    "basis_glow",
    "gen_perez_sky",
    "gen_grid",
    "parse_primitive",
    "parse_epw",
    "parse_wea",
    "parse_polygon",
    "Primitive",
    # "oconv",
    "points_as_sender",
    "rfluxmtx",
    # "rtrace",
    "sky_as_receiver",
    "surface_as_receiver",
    "sun_as_receiver",
    "surface_as_sender",
    # "render",
    "unpack_primitives",
    "View",
    "view_as_sender",
    "WeaData",
    "WeaMetaData",
]



"""Unit tests for frads.parsers module."""
import os
import sys
from pathlib import Path
sys.path.append(".")

from frads import parsers
from frads.geom import Vector


testdir = Path(os.path.dirname(__file__))
resdir = testdir / "Resources"
objdir = testdir / "Objects"
epw_path = resdir / "USA_CA_Oakland.Intl.AP.724930_TMY3.epw"
wea_path = resdir / "oak.wea"


def test_parse_mrad_config():
    window_paths = [
        Path("test", "Objects", "lower_glass.rad"),
        Path("test", "Objects", "upper_glass.rad"),
    ]
    vdict = {"vf": "test/v1a.vf", "x": 4, "y": 4}
    vmx_opt = {"ab": 2, "ad": 64, "lw": 1e-4}
    epw_path = Path("Resources", "USA_CA_Oakland.Intl.AP.724930_TMY3.epw")
    cfg_path = Path("test", "Resources", "test.cfg")
    config = parsers.parse_mrad_config(cfg_path)
    view = config["RaySender"].getview("view")
    assert config["Model"].getpaths("windows") == window_paths
    assert config["Model"].getpaths("ncps") == []
    assert view.vtype == "a"
    assert view.position == Vector(17, 5, 4)
    assert view.hori == 180
    assert config["SimControl"].getoptions("vmx_opt") == vmx_opt
    assert config["Site"].getpaths("epw_path") == [epw_path]
    assert config["Site"].getpaths("wea_path") is None


def test_parse_vu():
    inp_str = "-vta -vv 180 -vh 180 -vp 0 0 0 -vd 0 -1 0"
    res = parsers.parse_vu(inp_str)
    answer = {"vt": "a", "vv": 180, "vh": 180,
              "vp": [0, 0, 0], "vd": [0, -1, 0]}
    assert res.position == Vector(0, 0, 0)
    assert res.direction == Vector(0, -1, 0)
    assert res.vtype == "a"
    assert res.hori == 180
    res2 = parsers.parse_vu("")
    assert res2 is None


def test_parse_opt():
    inp_str = "-ab 8 -ad 1024 -I+ -u- -c 8 -aa .1 -lw 1e-8"
    res = parsers.parse_opt(inp_str)
    answer = {
        "ab": 8,
        "ad": 1024,
        "I": True,
        "u": False,
        "c": 8,
        "aa": 0.1,
        "lw": 1e-8,
    }
    assert res == answer


def test_parser_epw():
    with open(epw_path) as rdr:
        wea_metadata, wea_data = parsers.parse_epw(rdr.read())
    assert wea_data[0].time.month == 1
    assert wea_data[-1].time.month == 12
    assert wea_data[-1].time.day == 31
    assert wea_metadata.latitude == 37.72

def test_parser_wea():
    with open(wea_path) as rdr:
        wea_metadata, wea_data = parsers.parse_wea(rdr.read())
    assert wea_data[0].time.month == 1
    assert wea_data[-1].time.month == 12
    assert wea_data[-1].time.day == 31
    assert wea_data[-1].dni == 0
    assert wea_metadata.latitude == 37.72

def test_parse_idf():
    pass

def test_parse_optics():
    pass

def test_parse_igsdb_json():
    pass

def test_parse_branch():
    pass

def test_parse_ttree():
    pass

def get_nested_list_levels():
    pass



"""Call Radiance trace related stuff."""
import logging
from pathlib import Path
import subprocess as sp
from typing import Tuple
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence


from frads import sky
from frads import utils
from frads.types import Primitive
from frads.types import Options
from frads.types import View

logger = logging.getLogger("frads.raycall")


def get_oconv_command(
    *paths, has_stdin: bool = False, frozen: bool = False, octree=None
) -> List[str]:
    """Get oconv command.

    Args:
        paths: input file paths. (remember material definitions go first)
        has_stdin: whether we have stdin for this oconv call?
        frozen: Whether we want to have a frozen octree.
        octree: Existing octree to bind with.
    Returns:
        Oconv call command as a list of strings.
    """
    command = ["oconv"]
    if octree:
        command.append("-i")
        command.append("octree")
    if frozen:
        command.append("-f")
    if has_stdin:
        command.append("-")
    command.extend(paths)
    return command


def get_rtrace_command(
    option: List[str], octree: Path, header: bool = False
) -> List[str]:
    """Get rtrace command.

    Args:
        option: rtrace options.
        octree: octree path.
        header: Whether we want header info in the output.
    Returns:
        Rtrace call command as a list of strings.
    """
    command = ["rtrace"]
    if not header:
        command.append("-h")
    command += [*option, str(octree)]
    return command


def get_rfluxmtx_command(
    receiver: Path,
    option: Optional[List[str]] = None,
    sys_paths: Optional[Iterable[Path]] = None,
    octree: Optional[Path] = None,
    sender: Optional[Path] = None,
) -> List[str]:
    """
    Sender: stdin, polygon
    Receiver: surface with -o
    """
    command = ["rfluxmtx"]
    if option:
        command.extend(option)
    if isinstance(sender, Path):
        command.append(str(sender))
    else:
        command.append("-")
    command.append(str(receiver))
    if octree:
        command.extend(["-i", str(octree)])
    if sys_paths:
        for path in sys_paths:
            command.append(str(path))
    return command


def get_rcontrib_command(
    option: List[str], yres, outpath: Path, modifier: Path, octree: Path, xres=None
) -> List[str]:
    """
    Needs a wrapper for options.
    grouped by modifiers
    -n -V -c [ -fo | -r ]
    -e  -f
    -x -y -o -p -b -bn { -m | -M }
    [rtrace option]
    octree
    """
    command = ["rcontrib", *option]
    if xres:
        command += ["-ffc", "-x", xres]
    else:
        command.append("-faf")
    command += ["-y", yres]
    command += ["-o", str(outpath), "-M", str(modifier), str(octree)]
    command.append(str(octree))
    return command


def oconv(*paths, outpath, stdin=None, frozen: bool = False, octree=None) -> None:
    has_stdin = bool(stdin)
    oconv_command = get_oconv_command(
        *paths, has_stdin=has_stdin, octree=octree, frozen=frozen
    )
    logger.info(oconv_command)
    utils.run_write(oconv_command, outpath, stdin=stdin)


def rtrace(
    rays, option: List[str], octree: Path, out=None, header: bool = False
) -> Tuple[str, str]:
    rtrace_command = get_rtrace_command(option, octree, header=header)
    proc = sp.run(
        rtrace_command,
        check=True,
        encoding="ascii",
        input=rays,
        stderr=sp.PIPE,
        stdout=sp.PIPE,
    )
    return proc.stdout, proc.stderr


def pcond(
    hdr: Path,
    human: bool = False,
    acuity: bool = False,
    veiling: bool = False,
    sense: bool = False,
    closs: bool = False,
    linear: bool = False,
) -> str:
    """Condition a HDR image.

    This one has reduced functionality.

    Args:
        hdr: input hdr
        human: Mimic human visual response. This is the same as turning on all acuity,
            veiling, sensitivity, and color loss.
        acuity: Defocus darker region.
        veiling: Add veiling glare.
        sense: Use human contrast sensitivity, simulating eye internal scattering.
        linear: Use a linear reponse function instead of the standard dynamic range
            compression. This preseves the extremas.

    Returns:
        output hdr
    """
    command = ["pcond"]
    if human:
        command.append("-h")
    else:
        if acuity:
            command.append("-a")
        if veiling:
            command.append("-v")
        if sense:
            command.append("-s")
        if closs:
            command.append("-c")
    if linear:
        command.append("-l")
    command.append(str(hdr))
    proc = sp.run(command, check=True, stdout=sp.PIPE)
    return proc.stdout


def get_rpict_command(
    view: View,
    options: Optional[Options] = None,
    octree: Optional[Path] = None,
) -> List[str]:
    command = ["rpict"]
    command.extend(view.args())
    if options is not None:
        command.extend(options.args())
    if octree:
        command.append(str(octree))
    return command


def render(
    view: View,
    room: Sequence[Primitive],
    sky: Primitive,
    options: Optional[Options] = None,
):
    """Render a image."""
    oconv_command = get_oconv_command(has_stdin=True)
    input_prim = "\n".join(map(str, room))
    input_prim += "\n" + sky
    oconv_proc = sp.run(
        oconv_command,
        check=True,
        input=input_prim.encode(),
        stdout=sp.PIPE,
    )
    rpict_command = get_rpict_command(view, options)
    logger.info(rpict_command)
    rpict_proc = sp.run(
        rpict_command, check=True, input=oconv_proc.stdout, stdout=sp.PIPE
    )
    return rpict_proc


def renderf(*paths, view, wea, meta, options):
    """Render a image."""
    ssky = sky.gen_perez_sky(wea, meta)
    oconv_command = get_oconv_command(*paths, has_stdin=True)
    oconv_proc = sp.run(
        oconv_command, check=True, input=ssky.encode("ascii"), stdout=sp.PIPE
    )
    rpict_command = get_rpict_command(view, options)
    rpict_proc = sp.run(
        rpict_command, check=True, input=oconv_proc.stdout, stdout=sp.PIPE
    )
    return rpict_proc


def get_pixel_values(path=None, stdin=None, gamma: Optional[float] = None):
    """Get HDR pixel values.

    Use pvalue to get hdr pixel values as double precision floats.

    Args:
        path: hdr file path. Either path or stdin is used, path takes
            precedence.
        stdin: hdr data as bytes. Either path or stdin is used, path
            takes precedence.
    Returns:
        xres, yres, buf
    """
    cmd = ["pvalue", "-o", "-h", "-dd"]
    if gamma:
        cmd.append("-g")
        cmd.append(str(gamma))
    if path is not None:
        cmd.append(path)
        proc = sp.run(cmd, check=True, input=None, stdout=sp.PIPE)
    elif stdin:
        proc = sp.run(cmd, check=True, input=stdin, stdout=sp.PIPE)
    resstr, buf = proc.stdout.split(b"\n", 1)
    c1, yres, c2, xres = resstr.split()
    if c1 != b"-Y" and c2 != b"+X":
        raise ValueError("Non-standard Radiance image orientation.")
    return xres, yres, buf



import configparser
import os
from pathlib import Path
import sys
sys.path.append(".")

from frads import methods
from frads import parsers
from frads import utils
from frads.types import PaneProperty
from frads.types import PaneRGB
from frads.types import Primitive

test_dir_path = Path(__file__).parent.resolve()
reinsrc6_path = Path("Resources", "reinsrc6.rad")
grid_path = Path("Resources", "grid.pts")
prim_path = Path("Resources/model/Objects/floor_openroom.rad")

def test_assemble_model():
    os.chdir("test")
    config_path = Path("five_phase.cfg")
    config = parsers.parse_mrad_config(config_path)
    with methods.assemble_model(config) as model:
        assert os.path.isfile(model.material_path)
        assert len(model.window_groups) == 2
        assert len(model.bsdf_xml) == 2
        assert model.sender_grid['floor'] is not None
    os.chdir("..")



"""
Call uvspec to generate spectrally resolved sky description.
This module contains only functionalities relating to calling uvspec.
Uvspec needs to be in the local environment.
This module is not merged into sky.rad because of this unique circumstance.
"""

import argparse
from datetime import datetime, timedelta
import logging
import math
import os
from pathlib import Path
import subprocess as sp
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union

from frads import sky
from frads import parsers
from frads import color
from frads import color_data
from frads.types import Primitive
from frads.types import WeaMetaData


# Wavelength range and step
START_WVL = 360
END_WVL = 800
WVL_STEP = 10

# Approximate spectrum from 3 samples at 440, 550, and 660nm
# These coefficients are generated using apm_1nm.dat solar spectrum
# with alpha = 0, spectral range 360 - 800 nm
KR0, KG0, KB0 = 90487.185526, 68805.727959, 70873.299179

# These coefficiens are derived from CIE Colorimetry S0, S1, and S2 (Table T.2)
# integrated with cie_y_bar over 360 - 800 nm
YS0, YS1, YS2 = 10700.685138588558, 193.0296802325152, 76.46615567113169


def samples2spec(l_r, l_g, l_b) -> dict:
    """Approximate full spectrum from three samples."""
    c_r, c_g, c_b = l_r * KR0, l_g * KG0, l_b * KB0
    cie_x, cie_y, cie_z = color.rgb2xyz(c_r, c_g, c_b, color_data.RGB2XYZ_SRGB)
    chrom_x, chrom_y = color.xyz2xy(cie_x, cie_y, cie_z)
    scale_y = cie_y / color_data.LEMAX
    m1 = (-1.3515 - 1.7703 * chrom_x + 5.9114 * chrom_y) / (
        0.0241 + 0.2562 * chrom_x - 0.7341 * chrom_y
    )
    m2 = (0.0300 - 31.4424 * chrom_x + 30.0717 * chrom_y) / (
        0.0241 + 0.2562 * chrom_x - 0.7341 * chrom_y
    )
    spec = {
        wvl: scale_y * (_s[0] + _s[1] * m1 + _s[2] * m2) / (YS0 + YS1 * m1 + YS2 * m2)
        for wvl, _s in color_data.CIE_S012.items()
        if START_WVL <= wvl <= END_WVL
    }
    return spec


def get_uniform_samples(step: int) -> tuple:
    """Get uniform sky sampling angles.
    args:
        step: uniform sampling spacing (deg), need to be divisive by 90.
    return:
        cos(theta)
        phi angles
    """
    if 90 % step != 0:
        raise Exception("Angluar resolution not divisive by 90")
    thetas = range(0, 90, step)
    cos_thetas = [-math.cos(math.radians(i)) for i in thetas]
    phis = list(range(0, 361, step))
    return cos_thetas, phis


def get_solar(dt, meta) -> Tuple[Primitive, Sequence[Union[float, int]], float, float]:
    # year: str, month: str, day: str, hours: str, lat: str, lon: str, tzone: str
    # )
    """Call gendaylit to get solar angles."""
    hours = dt.hour + dt.minute / 60.0
    cmd = sky.gendaylit_cmd(
        dt.month,
        dt.day,
        hours,
        meta.latitude,
        meta.longitude,
        meta.timezone,
        year=dt.year,
    )
    gdl_proc = sp.run(list(map(str, cmd)), check=True, stdout=sp.PIPE, stderr=sp.PIPE)
    gdl_prims = parsers.parse_primitive(gdl_proc.stdout.decode().splitlines())
    source_prim = [prim for prim in gdl_prims if prim.ptype == "source"][0]
    source_dir = source_prim.real_arg[1:4]
    zenith_angle = math.degrees(math.acos(source_dir[2]))
    azimuth_angle = math.degrees(math.atan2(-source_dir[0], -source_dir[1])) % 360
    return source_prim, source_dir, zenith_angle, azimuth_angle


def gen_rad_template() -> str:
    """Generate sky.rad file template."""
    sky_template = "void colordata skyfunc\n"
    sky_template += (
        "9 noop noop noop {path}/red.dat {path}/green.dat {path}/blue.dat . "
    )
    sky_template += '"Acos(Dz)/DEGREE" "mod(atan2(-Dx, -Dy)/DEGREE,360)"\n'
    sky_template += "0\n0\n\n"
    sky_template += "skyfunc glow skyglow 0 0 4 1 1 1 0\n"
    sky_template += "skyglow source skydome 0 0 4 0 0 1 180\n"
    return sky_template


def gen_header(anglestep: int) -> str:
    """Generate header for colordata data files."""
    theta_interval = 90 / anglestep
    phi_interval = 360 / anglestep + 1
    header = "# Theta and phi dimensions\n2\n0 90 "
    header += f"{theta_interval:.0f}\n0 360 {phi_interval:.0f}\n"
    return header


def get_logger(verbosity: int) -> logging.Logger:
    """Setup logger.
    args:
        verbosity: verbosity levels 0-5
    returns:
        logger object
    """
    logger = logging.getLogger("frads.gencolorsky")
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    console_handler = logging.StreamHandler()
    _level = verbosity * 10
    logger.setLevel(_level)
    console_handler.setLevel(_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    return logger


def parse_cli_args() -> argparse.Namespace:
    """Parse commandline interface arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("year", type=int)
    parser.add_argument("month", type=int)
    parser.add_argument("day", type=int)
    parser.add_argument("hour", type=int)
    parser.add_argument("minute", type=int)
    parser.add_argument("-a", "--latitude", type=float, required=True)
    parser.add_argument("-o", "--longitude", type=float, required=True)
    parser.add_argument("-m", "--tzone", type=int, required=True)
    parser.add_argument("-e", "--atm")
    parser.add_argument(
        "-f", "--rgb", action="store_true", help="RGB sampling to full spectrum approx."
    )
    parser.add_argument("-s", "--observer", choices=["2", "10"], default="2")
    parser.add_argument("-c", "--colorspace", default="radiance")
    parser.add_argument("-b", "--cloudcover", type=float)
    parser.add_argument("-t", "--total", action="store_true")
    parser.add_argument("-g", "--cloudprofile", type=int)
    parser.add_argument("-r", "--anglestep", type=int, default=3)
    parser.add_argument("-u", "--altitude", default=0)
    parser.add_argument("-d", "--aod", type=float)
    parser.add_argument(
        "-l",
        "--aerosol",
        choices=[
            "continental_clean",
            "continental_average",
            "continental_polluted",
            "urban",
            "maritime_clean",
            "maritime_polluted",
            "maritime_tropical",
            "desert",
            "antarctic",
        ],
    )
    parser.add_argument("-i", "--pmt", action="store_true")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Verbose mode: \n"
        "\t-v=Debug\n"
        "\t-vv=Info\n"
        "\t-vvv=Warning\n"
        "\t-vvvv=Error\n"
        "\t-vvvvv=Critical\n"
        "default=Warning",
    )
    args = parser.parse_args()
    return args


class UvspecConfig:
    """Uvspec configuration.
    Attributes:
        aerosol: aerosol species file name (the file should reside in lib_path)
        albedo: surface reflectance
        altitude: altitude
        atmosphere: atmosphere composition profile
        azimuth: solar azimuth
        cloudcover: cloucover (0-1)
        latitude: latitude
        longitude: longitude
        output: user output setting (default: lambda ed edn uu)
        phis: Phi angles to sample
        solver: solver
        source: solar source file path
        time: date-time str (%Y %m %d %H %M %S)
        umu: cos(theta) angles to sample
        verbose: verbosity (verbose | quiet)
        wavelength: wavelength to sample
        zenith: solar zenith angle
    """

    def __init__(self):
        try:
            self.lib_path = Path(os.environ["LIBRADTRAN_DATA_FILES"])
        except KeyError as ke:
            raise "Can't find LIBRADTRAN_DATA_FILES in environment" from ke
        self._aerosol = ""
        self._albedo = 0.2
        self._altitude = 100
        self._aod = 0
        self._atmosphere = ""
        self._azimuth = 90
        self._band = ""
        self._cloudcover = 0
        self._latitude = "N 37"
        self._longitude = "W 122"
        self._output = "lambda edir edn uu"
        self._phis = ""
        self._post_process = ""
        self._solver = "pseudospherical"
        self._source = "apm_1nm"
        self._time = "2022 07 06 12 00 00"
        self._umu = ""
        self._verbose = "quiet"
        self._wavelength = 0
        self._zenith = 30
        self.wc_path = Path(__file__).parent / "data" / "WC.DAT"

    @property
    def config(self):
        """Get the uvpsec input string."""
        cfg = [f"data_files_path {str(self.lib_path)}"]
        cfg.append(f"source solar {str(self._source)}")
        if self._atmosphere:
            cfg.append(f"atmosphere_file {self._atmosphere}")
        cfg.append(self._solver)
        if self._band:
            cfg.append(f"mol_abs_param {self._band}")
        cfg.append(f"latitude {self._latitude}")
        cfg.append(f"longitude {self._longitude}")
        cfg.append(f"time {self._time}")
        cfg.append(f"altitude {self._altitude}")
        cfg.append(f"albedo {self._albedo}")
        cfg.append(f"sza {self._zenith}")
        cfg.append(f"phi0 {self._azimuth}")
        if self._cloudcover:
            cfg.append(f"wc_file 1D {self.wc_path}")
            cfg.append(f"cloudcover wc {self.cloudcover}")
            cfg.append("interpret_as_level wc")  # use independent pixel approximation
        if self._aerosol:
            cfg.append("aerosol_default")
            cfg.append(f"aerosol_species_file {self._aerosol}")
        if self._aod:
            cfg.append(f"aerosol_modify tau set {self._aod}")
        if self._umu:
            cfg.append(f"umu {self._umu}")
        if self._phis:
            cfg.append(f"phi {self._phis}")
        cfg.append(f"output_user {self._output}")
        if self._post_process:
            cfg.append(f"output_process {self._post_process}")
        cfg.append(self._verbose)
        if self._wavelength:
            cfg.append(f"wavelength {self._wavelength}")
        return "\n".join(cfg)

    @property
    def aerosol(self):
        """Get aerosol."""
        return self._azimuth

    @aerosol.setter
    def aerosol(self, path):
        self._aerosol = path

    @property
    def albedo(self):
        """Get ground albedo."""
        return self._albedo

    @albedo.setter
    def albedo(self, refl):
        self._albedo = refl

    @property
    def altitude(self):
        """Get site altitude."""
        return self._altitude

    @altitude.setter
    def altitude(self, alt):
        self._altitude = alt

    @property
    def aod(self):
        """Get aerosol optical depth."""
        return self._aod

    @aod.setter
    def aod(self, aod_):
        self._aod = aod_

    @property
    def atmosphere(self):
        """Get atmosphere file path."""
        return f"atmosphere_file {self._atmosphere}"

    @atmosphere.setter
    def atmosphere(self, atmosphere_file):
        self._atmosphere = atmosphere_file

    @property
    def azimuth(self):
        """Get solar azimuth angle."""
        return self._azimuth

    @azimuth.setter
    def azimuth(self, azimuth):
        self._azimuth = azimuth

    @property
    def band(self):
        """Get solar source band."""
        return self._band

    @band.setter
    def band(self, params):
        self._band = params

    @property
    def cloudcover(self):
        """Get cloudcover value."""
        return self._cloudcover

    @cloudcover.setter
    def cloudcover(self, cc):
        self._cloudcover = cc

    @property
    def latitude(self):
        """Get site latitude."""
        return self._latitude

    @latitude.setter
    def latitude(self, lat: float):
        if lat < 0:
            self._latitude = f"S {abs(lat):.2f}"
        else:
            self._latitude = f"N {lat:.2f}"

    @property
    def longitude(self):
        """Get site longitude."""
        return self._longitude

    @longitude.setter
    def longitude(self, lon: float):
        if lon < 0:
            self._longitude = f"E {abs(lon):.2f}"
        else:
            self._longitude = f"E {lon:.2f}"

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, ouput_user: str):
        self._output = ouput_user

    @property
    def post_process(self):
        return self._post_process

    @post_process.setter
    def post_process(self, process: str):
        self._post_process = process

    @property
    def phis(self):
        return self._phis

    @phis.setter
    def phis(self, phi: List[float]):
        self._phis = " ".join(map(str, phi))

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, solver):
        self._solver = solver

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source: str):
        self._source = source

    @property
    def time(self):
        return self._time

    @time.setter
    def time(self, dt_: str):
        self._time = dt_

    @property
    def umu(self):
        return self._umu

    @umu.setter
    def umu(self, ct: List[float]):
        self._umu = " ".join(map(str, ct))

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, v_q: str):
        if v_q.startswith("q"):
            self._verbose = "quiet"
        elif v_q.startswith("v"):
            self._verbose = "verbose"

    @property
    def wavelength(self):
        return self._wavelength

    @wavelength.setter
    def wavelength(self, wvl: float):
        self._wavelength = wvl

    @property
    def zenith(self):
        return self._zenith

    @zenith.setter
    def zenith(self, zenith: float):
        self._zenith = zenith


def genskylrt_total(config):
    """."""
    config.band = "kato2"
    config.output = "edir edn"
    config.post_process = "sum"
    proc = sp.run(
        "uvspec",
        encoding="utf-8",
        input=config.config,
        stderr=sp.PIPE,
        stdout=sp.PIPE,
    )
    direct_hor, diff_hor = list(map(float, proc.stdout.strip().split()))
    return direct_hor, diff_hor


def genskylrt_solar(config):
    """."""
    config.band = "kato2"
    config.post_process = "sum"
    proc = sp.run(
        "uvspec",
        encoding="utf-8",
        input=config.config,
        stderr=sp.PIPE,
        stdout=sp.PIPE,
    )
    return [i / 1e3 for i in map(float, proc.stdout.strip().split())]


def genskylrt(config: UvspecConfig, wavelengths):
    """."""
    result = []
    for wvl in wavelengths:
        config.wavelength = wvl
        proc = sp.run(
            "uvspec",
            encoding="utf-8",
            input=config.config,
            stdout=sp.PIPE,
        )
        result.append(proc.stdout.strip().split())
    return list(zip(*result))


def main() -> None:
    """gencolorsky entry point."""
    args = parse_cli_args()
    logger = get_logger(args.verbose)
    verbose = args.verbose == 1
    if args.rgb:
        wavelengths = [440, 550, 680]
    else:
        wavelengths = list(range(START_WVL, END_WVL + 1, WVL_STEP))
    dt = datetime(args.year, args.month, args.day, args.hour, args.minute)
    meta = WeaMetaData(
        "country", "city", args.latitude, args.longitude, args.tzone, args.altitude
    )
    ct, phis = get_uniform_samples(args.anglestep)
    direct_sun = True
    source_prim, source_dir, zenith_angle, azimuth_angle = get_solar(dt, meta)
    uvspec = UvspecConfig()
    uvspec.source = uvspec.lib_path / "solar_flux" / "apm_1nm"
    uvspec.solver = "pseudospherical"
    uvspec.time = (dt + timedelta(hours=int(meta.timezone / (-15)))).strftime(
        "%Y %m %d %H %M %S"
    )
    uvspec.latitude = meta.latitude
    uvspec.longitude = meta.longitude
    uvspec.altitude = meta.elevation
    uvspec.zenith = zenith_angle
    uvspec.azimuth = azimuth_angle
    uvspec.verbose = "verbose" if verbose else "quiet"
    if args.atm:
        uvspec.atmosphere = args.atm
    if args.cloudcover:
        if args.cloudprofile:
            uvspec.wc_path = args.cloudprofile
        uvspec.cloudcover = args.cloudcover
        direct_sun = args.cloudcover != 1.0
    if args.aerosol:
        uvspec.aerosol = args.aerosol
    uvspec.aod = args.aod
    if args.total:
        logger.info(uvspec.config)
        direct_hor, diff_hor = genskylrt_total(uvspec)
        direct_normal = direct_hor / source_dir[2]
        print(f"DNI: {direct_normal:.2f} W/m2")
        print(f"DHI: {diff_hor:.2f} W/m2")
        print(f"GHI: {direct_hor + diff_hor:.2f} W/m2")
        exit()
    uvspec.umu = ct
    uvspec.phis = phis
    uvspec.output = "lambda edir edn uu"
    if not direct_sun:
        uvspec.output = "lambda uu"
    logger.info(uvspec.config)
    columns = genskylrt(uvspec, wavelengths)
    if args.rgb:
        wavelengths = range(360, 801, 5)
        new_columns = [list(range(360, 801, 5))]
        for col in columns[1:]:
            new_columns.append(list(samples2spec(*map(float, col[::-1])).values()))
        columns = new_columns
    wvl_range = END_WVL - START_WVL + WVL_STEP
    cie_xyz_bar = color.get_interpolated_cie_xyz(wavelengths, args.observer)
    if args.pmt:
        blue = genskylrt_solar(uvspec)
        pfact = color_data.LEMAX * wvl_range / wvl_length / 1e3
        mfact = color_data.MLEMAX * wvl_range / wvl_length / 1e3
        red = []
        green = []
        mlnp = color.get_interpolated_mlnp(wavelengths)
        for col in columns[1:]:
            col = list(map(float, col))
            cieys = [i * j[1] for i, j in zip(col, cie_xyz_bar)]
            edis = [i * j for i, j in zip(col, mlnp)]
            cie_y = pfact * sum(cieys)
            edi = mfact * sum(edis)
            red.append(cie_y)
            green.append(edi)
    else:
        coeffs = color_data.XYZ2RGB_RAD
        pfact = color_data.LEMAX / 1e3 / 179
        red = []
        green = []
        blue = []
        for col in columns[1:]:
            col = list(map(float, col))
            cie_x, cie_y, cie_z = color.spec2xyz(cie_xyz_bar, col, wvl_range, emis=True)
            cie_x *= pfact
            cie_y *= pfact
            cie_z *= pfact
            _r, _g, _b = color.xyz2rgb(cie_x, cie_y, cie_z, coeffs)
            red.append(_r)
            green.append(_g)
            blue.append(_b)
    out_dir = Path(
        f"lrt_{args.month:02d}{args.day:02d}_{args.hour:02d}"
        f"{args.minute:02d}_{args.latitude}_{args.longitude}"
    )
    out_dir.mkdir(exist_ok=True)
    if direct_sun:
        sidx = 2
    else:
        sidx = 0
    header = gen_header(args.anglestep)
    with open(out_dir / "red.dat", "w") as wtr:
        wtr.write(header)
        wtr.write("\n".join([str(value) for value in red[sidx:]]))
    with open(out_dir / "green.dat", "w") as wtr:
        wtr.write(header)
        wtr.write("\n".join([str(value) for value in green[sidx:]]))
    with open(out_dir / "blue.dat", "w") as wtr:
        wtr.write(header)
        wtr.write("\n".join([str(value) for value in blue[sidx:]]))
    sky_template = gen_rad_template()
    with open(out_dir / "sky.rad", "w") as wtr:
        if direct_sun:
            wtr.write("void light solar\n0\n0\n3 ")
            wtr.write(f"{red[0]/sky.SOLAR_SA/source_dir[2]} ")
            wtr.write(f"{green[0]/sky.SOLAR_SA/source_dir[2]} ")
            wtr.write(f"{blue[0]/sky.SOLAR_SA/source_dir[2]}\n")
            wtr.write(str(source_prim) + "\n")
        wtr.write(sky_template.format(path=out_dir))



"""
This module contains routines to generate sender and receiver objects, generate
matrices by calling either rfluxmtx or rcontrib.
"""

from __future__ import annotations
import logging
import os
from pathlib import Path
import subprocess as sp
import tempfile as tf
from typing import Iterable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

from frads import sky
from frads import geom
from frads import parsers
from frads import raycall
from frads import utils
from frads.types import Primitive
from frads.types import Receiver
from frads.types import Sender
from frads.geom import Vector


logger: logging.Logger = logging.getLogger("frads.matrix")


def surface_as_sender(prim_list: list, basis: str, offset=None, left=None) -> Sender:
    """
    Construct a sender from a surface.

    Args:
        prim_list(list): a list of primitives
        basis(str): sender sampling basis
        offset(float): move the sender surface in its normal direction
        left(bool): Use left-hand rule instead for matrix generation

    Returns:
        A sender object (Sender)

    """
    prim_str = prepare_surface(
        prims=prim_list, basis=basis, offset=offset, left=left, source=None, out=None
    )
    logger.debug("Surface sender:\n%s", prim_str)
    return Sender("s", prim_str.encode(), None, None)


def view_as_sender(view: View, ray_cnt: int, xres: int, yres: int) -> Sender:
    """
    Construct a sender from a view.

    Args:
        view: a view object;
        ray_cnt: ray count;
        xres, yres: image resolution
        c2c: Set to True to trim the fisheye corner rays.

    Returns:
        A sender object

    """
    if (xres is None) or (yres is None):
        raise ValueError("Need to specify resolution")
    res_cmd = [
        "vwrays",
        *view.args(),
        "-x",
        str(xres),
        "-y",
        str(yres),
        "-d",
    ]
    logger.info("Check real image resolution: \n%s", " ".join(res_cmd))
    res_proc = sp.run(res_cmd, check=True, stdout=sp.PIPE, encoding="ascii")
    res_eval = res_proc.stdout.split()
    new_xres, new_yres = int(res_eval[1]), int(res_eval[3])
    if (new_xres != xres) and (new_yres != yres):
        logger.info("Changed resolution to %s %s", new_xres, new_yres)
    # vwrays_cmd = ["vwrays", "-ff", "-x", str(new_xres), "-y", str(new_yres)]
    vwrays_cmd = ["vwrays", "-ff"]
    if ray_cnt > 1:
        vwrays_cmd.extend(["-c", str(ray_cnt), "-pj", "0.7"])
    logger.debug("Ray count is %s", ray_cnt)
    vwrays_cmd += view.args()
    vwrays_cmd += ['-x', str(new_xres), '-y', str(new_yres)]
    logger.info("Generate view rays with: \n%s", " ".join(vwrays_cmd))
    vwrays_proc = sp.run(vwrays_cmd, check=True, stdout=sp.PIPE)
    if view.vtype == "a":
        flush_cmd = utils.get_flush_corner_rays_command(ray_cnt, xres)
        logger.info("Flushing -vta corner rays: \n%s", " ".join(flush_cmd))
        flush_proc = sp.run(
            flush_cmd, check=True, input=vwrays_proc.stdout, stdout=sp.PIPE
        )
        vrays = flush_proc.stdout
    else:
        vrays = vwrays_proc.stdout
    logger.debug("View sender:\n%s", vrays)
    return Sender("v", vrays, xres, yres)


def points_as_sender(pts_list: list, ray_cnt: Optional[int] = None) -> Sender:
    """
    Construct a sender from a list of points.

    Args:
        pts_list(list): a list of list of float
        ray_cnt(int): sender ray count

    Returns:
        A sender object
    """
    ray_cnt = 1 if ray_cnt is None else ray_cnt
    if pts_list is None:
        raise ValueError("pts_list is None")
    if not all(isinstance(item, list) for item in pts_list):
        raise ValueError("All grid points has to be lists.")
    pts_list = [i for i in pts_list for _ in range(ray_cnt)]
    grid_str = os.linesep.join([" ".join(map(str, li)) for li in pts_list]) + os.linesep
    logger.debug("Point sender:\n%s", grid_str)
    return Sender("p", grid_str.encode(), None, len(pts_list))


def sun_as_receiver(
    basis,
    smx_path: Path,
    window_normals: Optional[List[Vector]],
    full_mod: bool = False,
) -> Receiver:
    """
    Instantiate a sun receiver object.

    Args:
        basis: receiver sampling basis {kf | r1 | sc25...}
        smx_path: sky/sun matrix file path
        window_paths: window file paths
    Returns:
        A sun receiver object
    """

    # gensun = sky.Gensun(int(basis[-1]))
    if (smx_path is None) and (window_normals is None):
        str_repr, mod_str = sky.gen_sun_source_full(int(basis[-1]))
        return Receiver(str_repr, basis, modifier=mod_str)
    str_repr, mod_str, mod_str_full = sky.gen_sun_source_culled(
        int(basis[-1]), smx_path=smx_path, window_normals=window_normals
    )
    if full_mod:
        return Receiver(receiver=str_repr, basis=basis, modifier=mod_str_full)
    logger.debug("Sun receiver:\n%s", str_repr)
    logger.debug("Sun modifier:\n%s", mod_str)
    return Receiver(receiver=str_repr, basis=basis, modifier=mod_str)


def sky_as_receiver(basis: str, out) -> Receiver:
    """
    Instantiate a sky receiver object.

    Args:
        basis: receiver sampling basis {kf | r1 | sc25...}
    Returns:
        A sky receiver object
    """

    if not basis.startswith("r"):
        raise ValueError(f"Sky basis need to be Treganza/Reinhart, found {basis}")
    out.parent.mkdir(exist_ok=True)
    sky_str = f'#@rfluxmtx o="{str(out)}"\n'
    sky_str += sky.basis_glow(basis)
    logger.debug("Sky receiver:\n%s", sky_str)
    return Receiver(sky_str, basis)


def surface_as_receiver(
    prim_list: Sequence[Primitive],
    basis: str,
    out: Union[None, str, Path],
    offset=None,
    left: bool = False,
    source: str = "glow",
) -> Receiver:
    """
    Instantiate a surface receiver object.

    Args:
        prim_list: list of primitives(dict)
        basis: receiver sampling basis {kf | r1 | sc25...}
        out: output path
        offset: offset the surface in its normal direction
        left: use instead left-hand rule for matrix generation
        source: light source for receiver object {glow|light}
    Returns:
        A surface receiver object
    """
    rcvr_str = prepare_surface(
        prims=prim_list, basis=basis, offset=offset, left=left, source=source, out=out
    )
    logger.debug("Surface receiver:\n%s", rcvr_str)
    return Receiver(rcvr_str, basis)


def prepare_surface(*, prims, basis, left, offset, source, out) -> str:
    """
    Prepare the sender or receiver surface, adding appropriate tags.

    Args:
        prims(list): list of primitives
        basis(str): sampling basis
        left(bool): use instead the left-hand rule
        offset(float): offset surface in its normal direction
        source(str): surface light source for receiver
        out: output path
    Returns:
        The surface sender/receiver primitive as string
    """

    if basis is None:
        raise ValueError("Sampling basis cannot be None")
    if (source is not None) and (source not in ("glow", "light")):
        raise ValueError(f"Unknown source type {source}")
    upvector = utils.up_vector(prims)
    upvector = upvector.scale(-1) if left else upvector
    upvector_str = str(upvector).replace(" ", ",")
    modifier_set = {p.modifier for p in prims}
    if len(modifier_set) != 1:
        logger.warning("Primitives don't share modifier")
    src_mod = f"rflx{prims[0].modifier}{utils.id_generator()}"
    header = f"#@rfluxmtx h={basis} u={upvector_str}\n"
    if out is not None:
        header += f'#@rfluxmtx o="{out}"\n\n'
    if source == "glow":
        source_prim = Primitive("void", source, src_mod, ("0"), (4, 1, 1, 1, 0))
        header += str(source_prim)
    elif source == "light":
        source_prim = Primitive("void", source, src_mod, ("0"), (3, 1, 1, 1))
        header += str(source_prim)
    modifiers = [p.modifier for p in prims]
    content = ""
    for prim in prims:
        if prim.identifier in modifiers:
            _identifier = "discarded"
        else:
            _identifier = prim.identifier
        _modifier = src_mod
        if offset is not None:
            poly = parsers.parse_polygon(prim.real_arg)
            offset_vec = poly.normal.scale(offset)
            moved_pts = [pt + offset_vec for pt in poly.vertices]
            _real_args = geom.Polygon(moved_pts).to_real()
        else:
            _real_args = prim.real_arg
        new_prim = Primitive(
            _modifier, prim.ptype, _identifier, prim.str_arg, _real_args
        )
        content += str(new_prim) + "\n"
    return header + content


def rfluxmtx(
    sender: Sender,
    receiver: Receiver,
    env: Iterable[Path],
    opt: Optional[List[str]] = None,
) -> None:
    """
    Calling rfluxmtx to generate the matrices.

    Args:
        sender: Sender object
        receiver: Receiver object
        env: model environment, basically anything that's not the
            sender or receiver
        opt: option string

    Returns:
        return the stdout of the rfluxmtx run.
    """
    if None in (sender, receiver):
        raise ValueError("Sender/Receiver object is None")
    opt = [] if opt is None else opt
    _sender = None
    stdin: Optional[bytes] = sender.sender
    with tf.TemporaryDirectory() as tempd:
        receiver_path = Path(tempd, "receiver")
        with open(receiver_path, "w", encoding="ascii") as wtr:
            wtr.write(receiver.receiver)
        if sender.form == "s":
            sender_path = Path(tempd, "sender")
            with open(sender_path, "wb") as wtr:
                wtr.write(sender.sender)
            _sender = sender_path
            stdin = None
        elif sender.form == "p":
            opt.extend(["-I+", "-faa", "-y", str(sender.yres)])
        elif sender.form == "v":
            opt.extend(["-ffc", "-x", str(sender.xres), "-y", str(sender.yres), "-ld-"])
        cmd = raycall.get_rfluxmtx_command(
            receiver_path, option=opt, sender=_sender, sys_paths=env
        )
        logger.info("Running rfluxmtx with:\n%s", " ".join(cmd))
        try:
            sp.run(cmd, check=True, input=stdin, stderr=sp.PIPE)
        except sp.CalledProcessError as err:
            logger.error("rfluxmtx failed with:\n%s", err.stderr.decode("ascii"))
            raise err


def rcvr_oct(receiver, env, oct_path: Union[str, Path]) -> None:
    """
    Generate an octree of the environment and the receiver.

    Args:
        receiver: receiver object
        env: environment file paths
        oct_path: Path to write the octree to
    Returns:
        None
    """

    with tf.TemporaryDirectory() as tempd:
        receiver_path = os.path.join(tempd, "rcvr_path")
        with open(receiver_path, "w", encoding="utf-8") as wtr:
            wtr.write(receiver.receiver)
        ocmd = ["oconv", "-f", *map(str, env), receiver_path]
        logger.info("Generate octree with:\n%s", " ".join(ocmd))
        with open(oct_path, "wb") as wtr:
            sp.run(ocmd, check=True, stdout=wtr)


def rcontrib(
    sender,
    modifier: str,
    octree: Union[str, Path],
    out: Union[str, Path],
    opt: List[str],
) -> None:
    """
    Calling rcontrib to generate the matrices.

    Args:
        sender: Sender object
        modifier: modifier str listing the receivers in octree
        octree: the octree that includes the environment and the receiver
        opt: option string
        out: output path

    Returns:
        None

    """
    opt.append("-fo+")
    with tf.TemporaryDirectory() as tempd:
        modifier_path = os.path.join(tempd, "modifier")
        with open(modifier_path, "w", encoding="utf-8") as wtr:
            wtr.write(modifier)
        cmd = ["rcontrib", *opt]
        stdin = sender.sender
        if sender.form == "p":
            cmd += ["-I+", "-faf", "-y", str(sender.yres)]
        elif sender.form == "v":
            out = Path(out)
            out.mkdir(exist_ok=True)
            out = out / "%04d.hdr"
            cmd += ["-ffc", "-x", str(sender.xres), "-y", str(sender.yres)]
        cmd += ["-o", str(out), "-M", modifier_path, str(octree)]
        logger.info("Running rcontrib with:\n%s", " ".join(cmd))
        sp.run(cmd, check=True, input=stdin)



import os
from pathlib import Path
import sys
sys.path.append('.')
from frads import geom


def test_merge_window():
    """
    0,0,7-------3,0,7   4,0,7-------8,0,7
      |           |       |           |
      |           |       |           |
      |           |       |           |
    0,0,4-------3,0,4   4,0,4-------8,0,4

    0,0,3-------3,0,3   4,0,3-------8,0,3
      |           |       |           |
      |           |       |           |
      |           |       |           |
    0,0,0-------3,0,0   4,0,0-------8,0,0
    """
    pt1 = geom.Vector(0, 0, 0)
    pt2 = geom.Vector(3, 0, 0)
    pt3 = geom.Vector(4, 0, 0)
    pt4 = geom.Vector(8, 0, 0)
    pt5 = geom.Vector(0, 0, 3)
    pt6 = geom.Vector(3, 0, 3)
    pt7 = geom.Vector(4, 0, 3)
    pt8 = geom.Vector(8, 0, 3)
    pt9 = geom.Vector(0, 0, 4)
    pt10 = geom.Vector(3, 0, 4)
    pt11 = geom.Vector(4, 0, 4)
    pt12 = geom.Vector(8, 0, 4)
    pt13 = geom.Vector(0, 0, 7)
    pt14 = geom.Vector(3, 0, 7)
    pt15 = geom.Vector(4, 0, 7)
    pt16 = geom.Vector(8, 0, 7)

    plg1 = geom.Polygon([pt1, pt5, pt6, pt2])
    plg2 = geom.Polygon([pt3, pt7, pt8, pt4])
    plg3 = geom.Polygon([pt9, pt13, pt14, pt10])
    plg4 = geom.Polygon([pt11, pt15, pt16, pt12])

    answer12 = geom.Polygon([pt1, pt5, pt8, pt4])
    answer123 = geom.Polygon([pt1, pt13, pt14, pt8, pt4])
    answer1234 = geom.Polygon([pt1, pt13, pt16, pt4])

    res12 = geom.merge_polygon([plg1, plg2])
    res123 = geom.merge_polygon([plg1, plg2, plg3])
    res1234 = geom.merge_polygon([plg1, plg2, plg3, plg4])

    assert res12 == answer12
    assert res123 == answer123
    assert res1234 == answer1234




import os
from pathlib import Path
import unittest
from frads import mtxmult


class TestMtxmult(unittest.TestCase):

    test_dir_path = os.path.dirname(__file__)
    data_path = os.path.join(test_dir_path, "Resources")
    check_decimal_to = 6

    def test_pcomb(self):
        # inp =
        # ops =
        # out_dir =
        # nproc = 1
        # mtxmult.pcomb(inp, ops, out_dir, nproc=nproc)
        # self.assertEqual(res, answer)
        # nproc = 4
        # mtxmult.pcomb(inp, ops, out_dir, nproc=nproc)
        # self.assertEqual(res, answer)
        pass

    def test_mtxmult(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""
This module contains all color and spectral-related functionalities.
"""
from typing import List
from typing import Sequence
from typing import Tuple
from typing import Union

from frads import color_data
from frads.types import ColorPrimaries


def get_interpolated_cie_xyz(
    inp_wvl: Sequence[Union[float, int]], observer: str
) -> List[tuple]:
    """Load CIE tristimulus data according to input wavelength.
    Also load melanopic action spectra data as well.

    Args:
        inp_wvl: a list of input wavelength in nm
        observer: 2° or 10° observer for the colar matching function.
    Returns:
        CIE-x
        CIE-y
        CIE-z
        Melanopic action spectra
        Index of input wavelength corresponding to available tristimulus data
    """

    if observer.startswith("2"):
        cie_xyz = color_data.CIE_XYZ_2
    elif observer.startswith("10"):
        cie_xyz = color_data.CIE_XYZ_10
    common_wvl = sorted(set(cie_xyz).intersection(inp_wvl))
    return [cie_xyz[wvl] for wvl in common_wvl]


def get_interpolated_mlnp(
    inp_wvl: Sequence[Union[float, int]],
) -> List[float]:
    """Load CIE tristimulus data according to input wavelength.
    Also load melanopic action spectra data as well.

    Args:
        inp_wvl: a list of input wavelength in nm
    Returns:
        Melanopic action spectra
    """
    mlnp = color_data.CIE_MLNP
    common_wvl = sorted(set(mlnp).intersection(inp_wvl))
    return [mlnp[wvl] for wvl in common_wvl]


def get_conversion_matrix(prims: ColorPrimaries, reverse: bool = False) -> tuple:
    """
    Get CIE conversion matrix based on color primaries.

    Args:
        prims: Color space color primaries
        reverse: get RGB to XYZ conversion matrix instead
    Returns:
        The conversion matrix coefficients in a 1-dimensional tuple.
    """

    yw_inv = 1 / prims.yw

    cie_d = (
        prims.xr * (prims.yg - prims.yb)
        + prims.xg * (prims.yb - prims.yr)
        + prims.xb * (prims.yr - prims.yg)
    )
    cie_c_rd = yw_inv * (
        prims.xw * (prims.yg - prims.yb)
        - prims.yw * (prims.xg - prims.xb)
        + prims.xg * prims.yb
        - prims.xb * prims.yg
    )
    cie_c_gd = yw_inv * (
        prims.xw * (prims.yb - prims.yr)
        - prims.yw * (prims.xb - prims.xr)
        - prims.xr * prims.yb
        + prims.xb * prims.yr
    )
    cie_c_bd = yw_inv * (
        prims.xw * (prims.yr - prims.yg)
        - prims.yw * (prims.xr - prims.xg)
        + prims.xr * prims.yg
        - prims.xg * prims.yr
    )

    coeff_00 = (
        prims.yg - prims.yb - prims.xb * prims.yg + prims.yb * prims.xg
    ) / cie_c_rd
    coeff_01 = (
        prims.xb - prims.xg - prims.xb * prims.yg + prims.xg * prims.yb
    ) / cie_c_rd
    coeff_02 = (prims.xg * prims.yb - prims.xb * prims.yg) / cie_c_rd
    coeff_10 = (
        prims.yb - prims.yr - prims.yb * prims.xr + prims.yr * prims.xb
    ) / cie_c_gd
    coeff_11 = (
        prims.xr - prims.xb - prims.xr * prims.yb + prims.xb * prims.yr
    ) / cie_c_gd
    coeff_12 = (prims.xb * prims.yr - prims.xr * prims.yb) / cie_c_gd
    coeff_20 = (
        prims.yr - prims.yg - prims.yr * prims.xg + prims.yg * prims.xr
    ) / cie_c_bd
    coeff_21 = (
        prims.xg - prims.xr - prims.xg * prims.yr + prims.xr * prims.yg
    ) / cie_c_bd
    coeff_22 = (prims.xr * prims.yg - prims.xg * prims.yr) / cie_c_bd

    if reverse:
        coeff_00 = prims.xr * cie_c_rd / cie_d
        coeff_01 = prims.xg * cie_c_gd / cie_d
        coeff_02 = prims.xb * cie_c_bd / cie_d
        coeff_10 = prims.yr * cie_c_rd / cie_d
        coeff_11 = prims.yg * cie_c_gd / cie_d
        coeff_12 = prims.yb * cie_c_bd / cie_d
        coeff_20 = (1 - prims.xr - prims.yr) * cie_c_rd / cie_d
        coeff_21 = (1 - prims.xg - prims.yg) * cie_c_gd / cie_d
        coeff_22 = (1 - prims.xb - prims.yb) * cie_c_bd / cie_d

    return (
        coeff_00,
        coeff_01,
        coeff_02,
        coeff_10,
        coeff_11,
        coeff_12,
        coeff_20,
        coeff_21,
        coeff_22,
    )


def rgb2xyz(
    red: float, green: float, blue: float, coeffs: tuple
) -> Tuple[float, float, float]:
    """Convert RGB to CIE XYZ.

    Args:
        red: red
        green: green
        blue: blue
        coeffs: coversion matrix.
    Returns:
        CIE X, Y, Z.

    Raise:
        ValueError with invalid coeffs.
    """
    if len(coeffs) != 9:
        raise ValueError(f"{len(coeffs)} coefficients found, expected 9")
    ciex = coeffs[0] * red + coeffs[1] * green + coeffs[2] * blue
    ciey = coeffs[3] * red + coeffs[4] * green + coeffs[5] * blue
    ciez = coeffs[6] * red + coeffs[7] * green + coeffs[8] * blue
    return ciex, ciey, ciez


def xyz2rgb(
    cie_x: float, cie_y: float, cie_z: float, coeffs: tuple
) -> Tuple[float, float, float]:
    """Convert CIE XYZ to RGB.

    Args:
        cie_x: cie_x
        cie_y: cie_y
        cie_z: cie_z
        coeffs: conversion matrix
    Returns:
        Red, Green, Blue

    Raise:
        ValueError for invalid coeffs.
    """
    if len(coeffs) != 9:
        raise ValueError(f"{len(coeffs)} coefficients found, expected 9")
    red = max(0, coeffs[0] * cie_x + coeffs[1] * cie_y + coeffs[2] * cie_z)
    green = max(0, coeffs[3] * cie_x + coeffs[4] * cie_y + coeffs[5] * cie_z)
    blue = max(0, coeffs[6] * cie_x + coeffs[7] * cie_y + coeffs[8] * cie_z)
    return red, green, blue


def spec2xyz(
    cie_xyz_bar,
    spec: list,
    wvl_range: float,
    emis: bool = False,
) -> tuple:
    """Convert spectral data to CIE XYZ.

    Args:
        cie_xyz_bar: CIE color matching function.
        spec: input spectral data as a dictionary sorted by wavelenth as key.
        emis: flag whether the input data is emissive in nature.
    Returns:
        CIE X, Y, Z

    Note:
        Assuming input wavelength is a subset of the CIE ones
    """
    sval_cie = [
        (sv * cxyz[0], sv * cxyz[1], sv * cxyz[2])
        for sv, cxyz in zip(spec, cie_xyz_bar)
    ]
    spec_len = len(spec)
    cie_x, cie_y, cie_z = [sum(s_c) / spec_len * wvl_range for s_c in zip(*sval_cie)]
    if not emis:
        avg_y = sum(i[1] for i in cie_xyz_bar) / spec_len * wvl_range
        cie_x /= avg_y
        cie_y /= avg_y
        cie_z /= avg_y
    return cie_x, cie_y, cie_z


def xyz2xy(cie_x: float, cie_y: float, cie_z: float) -> Tuple[float, float]:
    """Convert CIE XYZ to xy chromaticity.

    Args:
        cie_x: CIE X
        cie_y: CIE Y
        cie_z: CIE Z
    Returns:
        x chromoticity
        y chromoticity
    """
    _sum = cie_x + cie_y + cie_z
    if _sum == 0:
        return 0, 0
    chrom_x = cie_x / _sum
    chrom_y = cie_y / _sum
    return chrom_x, chrom_y
